/**
 * @file liste.h
 * @Author Markus Gasteiger
 * @date January, 2018
 * @brief A collection of classes and functions to facilitate compile time handling of data.
 *
 */

#ifndef liste_h
#define liste_h

#include <cmath>
#include <type_traits>

#include "helpers.h"
#include "debug.h"

/**
* @brief Liste is a container to a tuple of compile time values.
* @tparam V is type of the list entries.
* @tparam names... are the entries of the list.
*
*/
template<typename V = char, V ... names>
struct liste;
/**
* @brief The main part of the template structure.
* @tparam V is the type of the list.
* @tparam name is the first entry in the list and
* @tparam names... are the remaining entries.
*
*/
template<typename V, V name, V ... names>
struct liste<V,name,names...>{
    /**
    * @brief The type of this list.
    *
    */
    using type = V;
    /**
    * @brief The length of the list.
    *
    */
    constexpr const static size_t size = 1 + liste<V,names...>::size;
    /**
    * @brief std::array access to the list.
    *
    */
    constexpr const static std::array<V,size> list = { name, names ... };

    /**
    * @brief The constructor is empty, because all information is based upon compile time parameters, i.e. templates.
    *
    */
    liste(){};

    /**
    * @brief The first member of the list is called head.
    *
    */
    constexpr static V head = name;
    /**
    * @brief The remaining terms are the rest.
    *
    */
    using rest = liste<V,names...>;

    /**
    * @brief Generate a new list with a value appended to the end of the current list.
    * @tparam var is the value to append.
    *
    */
    template<V var>
    using append=liste<V,names...,var>;

    /**
    * @brief ...
    *
    * @tparam var value to find in list.
    * @param pos is the offset to the position.
    * @return returns the positions or -1, if the value could not be found in the list.
    */
    template<V var, int pos=0>
    constexpr static int find_in(){
        if constexpr (name == var){ return pos;};
        return liste<V,names...>::template find_in<var,(pos+1)>();
    }

    /**
    * @brief Print the list to debug.
    *
    */
    constexpr static void debug(){
        debug::msg2 << name << " ";
        liste<V,names...>::debug();
    }

    /**
    * @brief Check that there are no duplicate entries in the list.
    *
    * @return Returns true if the list is unique, otherwise false.
    */
    constexpr static bool is_uniq(){
        if constexpr (liste<V,names...>::template find_in<name>() >=0){return false;}
        return liste<V,names...>::is_uniq();
    }

    template<class sub_liste>
    constexpr static std::array<size_t,sub_liste::size> sub_dim_pos(){
        static_assert(sub_liste::is_uniq());

        std::array<size_t,sub_liste::size> output {};
        size_t i=0;
        size_t j=0;
        for(; i<sub_liste::size; ++i){
            for(; j<size; ++j){
                if(sub_liste::list[i]==list[j]){
                    output[i]=j;
                    break;
                }
            }
            assert(j<size && "The sub_liste must be a true subspace.");
        }
        return output;
    }

    /**
    * @brief Get the value at a specific list position.
    *
    * @tparam pos is the position one wants the value to.
    * @return V
    */
    template<int pos>
    constexpr static V value(){
        static_assert(pos<size,"List is too small.");
        static_assert(pos>=0,"No negative positions.");
        if constexpr (pos == 0){
            return name;
        } else {
            return liste<V,names...>::template value<pos-1>();
        }
    }

    /**
    * @brief The run-time equivalent to value. It just returns the pos value of the list.
    *
    * @param pos is the position of the value to return.
    * @return V
    */
    inline V operator[](int pos) const { return list[pos];}

//     template<typename NEW, NEW val, NEW ... lst>
//     using new_val = typename liste<V,names...>::template new_val<NEW, val, lst..., val>;


    /**
    * @brief Inner product. It is meant to be used with strides.
    *
    * @tparam IN is the underlying type of the input std::array.
    * @param _in is the the input std::array.
    * @return V
    *
    */
    template<typename IN>
    constexpr static inline V inner_product(const std::array<IN,size>& _in) {
        IN output = 0;
        for(size_t i=0; i<size; ++i){
            output += _in[i]* (IN) list[i];
        }
        return output;
    }

/**
* @brief This function is specific to strides, which can also be stored as a list. It computes the corresponding coordinates from the linear strided position.
*
* @param pos is the linear position.
* @return coordinates.
*/
constexpr static inline std::array<V,size> pos2coords(V pos) {
        std::array<V,size> output {};
        for(int i = size-1; i>=0; --i){
            assert(list[i]!=0);
            output[i] = (pos/list[i]);
            pos = pos % list[i];
        }
        return output;
    }

template<class new_names>
constexpr static inline std::array<bool,size> mask(){
    std::array<bool,size> output {};
    size_t j=0;
    for(size_t i=0; i<size; ++i){
        for(j=0; j<new_names::size; ++j){
            if(list[i]==new_names::list[j]){
                output[i]=true;
            }
            if(j==new_names::size){
                output[i]=false;
            }
        }
    }
    return output;
}
};

template<int, typename, class>
struct listval;
template<typename V, V vl, V... vals>
struct listval<0,V,liste<V,vl,vals...>>{
    constexpr static const V val = vl;
};
template<int ii, typename V, V vl, V... vals>
struct listval<ii,V,liste<V,vl,vals...>>{
    constexpr static const V val = listval<(ii-1),V,liste<V,vals...>>::val;
};

template<typename V>
struct liste<V>{
    using type = V;
    using end = void;

    constexpr const static int size = 0;
    constexpr const static std::array<V,0> list= { };

    template<V var>
    using append=liste<V,var>;

    constexpr static void debug(){
        debug::msg2 << std::endl;
    }

template<int i=0, int ... is>
constexpr const static auto iters = std::array<int,i+1>({is...,i});
template<int i=0, int ... is>
constexpr const static auto reverse_iters = std::array<int,i+1>({is...,0});

    template<V M, V ... product>
    using multiply = liste<V,product...>;

    template<V var, int pos=0>
    constexpr static int find_in(){
        return -1;
    }

    template<typename NEW, NEW val, NEW ... lst>
    using new_val = liste<NEW,lst...>;

    template<int pos, V new_val, V... vals>
    using liste_set = liste<V,vals...>;

    constexpr static bool is_uniq(){
        return true;
    }

};

template<typename V, V ... names>
V operator*(const std::array<V, liste<V,names...>::size>& _in, const liste<V,names...>& _l ){
    return _l.inner_product(_in);
}
template<typename V, V ... names>
V operator*(const liste<V,names...>& _l, const std::array<V, liste<V,names...>::size>& _in){
    return _l.inner_product(_in);
}
template<typename V, V ... names>
std::array<V, liste<V,names...>::size> operator%(V _in, const liste<V,names...>& _l ){
    return _l.pos2coords(_in);
}

template<typename V, int length, V val, V ... vals>
struct liste_{
    using fill = typename liste_<V,length-1,val,vals...,val>::fill;
};
template<typename V, V val, V ... vals>
struct liste_<V,0,val,vals...>{
    using fill = liste<V,vals...>;
};

template<typename V, V ...>
struct make_strides;
template<typename V, V val, V ... vals>
struct make_strides<V, val, vals...>{
    template<V e=1, V ... z>
    using list=typename make_strides<V,vals ... >::template list<e*val, z..., e>;

    template<V... lvals>
    using rev_list = typename make_strides<V, vals...>::template rev_list<lvals..., (...*vals) >;
};
template<typename V, V val>
struct make_strides<V, val>{
    template<V e, V ... z>
    using list=liste<V, z..., e>;

    template<V ... lvals>
    using rev_list = liste<V, lvals..., 1>;
};

template<class>
struct liste2strides;
template<size_t val, size_t ... vals>
struct liste2strides<liste<size_t,val, vals...> >{
    template<size_t e=1, size_t ... z>
    using list=typename make_strides<size_t,vals ... >::template list<e*val, z..., e>;
    constexpr static size_t size = (...*vals)*val;
};
template<size_t val>
struct liste2strides<liste<size_t,val>>{
    constexpr static size_t size = val;
    template<size_t e, size_t ... z>
    using list=liste<size_t, z..., e>;
};
template<>
struct liste2strides<liste<size_t>>{
constexpr static size_t size = 0;

};

template<typename V, class l1, class l2>
struct merge_;
template<typename V, V ... l1, V ...l2>
struct merge_<V, liste<V,l1...>, liste<V,l2...> >{
    using list = liste<V,l1...,l2...>;
};

template<typename V, class l1, class l2, V ...>
struct lists_add;
template<typename V, V L1, V ... l1, V L2, V ...l2, V ... r1>
struct lists_add<V, liste<V,L1,l1...>, liste<V,L2,l2...>, r1... >{
    using reslt = typename lists_add<V,liste<V,l1...>,liste<V,l2...>,r1...,(L1+L2)>::reslt;
};
template<typename V, V ... r1>
struct lists_add<V, liste<V>, liste<V>, r1... >{
    using reslt = liste<V,r1...>;
};

#endif
