/**
 * @file debug.h
 * @Author Markus Gasteiger
 * @date September, 2017
 * @brief A slim debugging interface is defined. The log level specific streambuffers debug::msg{1..4} are provided and can be used for debugging.
 *
 */

#ifndef debug_h
#define debug_h

#include <string.h>
#include <iostream>
#include <ostream>

//LOCTMP is meant to elevate a local debug output temporarily and therefor has always to be the highest.
/** @enum mapper::debug_level
*   @brief The log levels are
*
*/
enum class debug_level {
    LOCTMP=-1, /**< this level is not for the global setting, but to use locally in the code to force messages */
    NO=0, /**< no logging */
    MINIMUM=1, /**< only really necessary message */
    HINT=2, /**< additional messages (to hint on possible improvements) */
    FULL=3, /**< all messages possible are shown */
};

/*!
* @brief The functionality for the debugging interface is collected in this namespace.
*
*/
namespace debug{
    /**
    * @brief The global debug level should be set at the beginning.
    *
    */
    debug_level level=debug_level::HINT;
    /**
    * @brief The default output target is clog, but can be changed at any time.
    *
    */
    std::ostream &output=std::clog;


    // suggested at https://stackoverflow.com/questions/7818371/printing-to-nowhere-with-ostream
    /// \todo verify the efficiency of these classes and whether they are cleaned by the compiler.
    class null_out_buf : public std::streambuf {
    public:
        virtual std::streamsize xsputn (const char * s, std::streamsize n) {
            return n;
        }
        virtual int overflow (int c) {
            return 1;
        }
    };
    class null_out_stream : public std::ostream {
    public:
        null_out_stream() : std::ostream (&buf) {}
    private:
        null_out_buf buf;
    };
    null_out_stream cnul;
//     shotty alternative:
//     std::ostream cnul(0);

    /**
    * @brief This function returns a streambuffer for the given debug level and already inserts a line header.
    * The stream buffer is the globally set output or the null interface (if there is no debugging).
    *
    * @param dbg_lvl the minimum debug level for which this message will be printed
    * @param file a source file name
    * @param line the line number in given source file naem.
    * @return std::ostream&
    */
    std::ostream &message(debug_level dbg_lvl=debug_level::HINT, const char* file=__FILE__, int line=__LINE__){
#ifndef NDEBUG
        if(dbg_lvl<=level){
            output << "" << file << ":" << line << ": ";
            return output;
        }
#endif
        return cnul;
    }
    /**
     * @brief This function returns a streambuffer for the given debug level and already inserts a line header.
     * The stream buffer is the globally set output or the null interface (if there is no debugging).
     *
     * @tparam dbg_lvl the minimum debug level for which this message will be printed
     * @param file a source file name
     * @param line the line number in given source file naem.
     * @return std::ostream&
     */
template<debug_level dbg_lvl=debug_level::HINT>
    std::ostream &message( const char* file=__FILE__, int line=__LINE__){
#ifndef NDEBUG
        if(dbg_lvl<=level){
            output << "" << file << ":" << line << ": ";
            return output;
        }
#endif
        return cnul;
    }


#define __FILENAME__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)
/**@def msg
    * @brief General wrapper around the message function that already fills file and line information.
    *
    */
#define msg(dbg_lvl) message<dbg_lvl>( __FILENAME__, __LINE__)

/**@def msg1
     * @brief Specific wrapper around the message function at debug_level::MINIMUM that already fills file and line information.
    *
    */
#define msg1 message<debug_level::MINIMUM>( __FILENAME__, __LINE__)

/**@def msg2
 * @brief Specific wrapper around the message function at debug_level::HINT that already fills file and line information.
*
*/
#define msg2 message<debug_level::HINT>( __FILENAME__, __LINE__)
/**@def msg3
 * @brief Specific wrapper around the message function at debug_level::FULL that already fills file and line information.
*
*/
#define msg3 message<debug_level::FULL>( __FILENAME__, __LINE__)
/**@def msg4
 * @brief Specific wrapper around the message function at debug_level::LOCTMP that already fills file and line information.
*
*/
#define msgf message<debug_level::LOCTMP>( __FILENAME__, __LINE__)


// #define warn(msg) {\
// #warn msg\
// output << msg << std::endl;\
// }


}
#endif
