/**
 * @file axis.h
 * @Author Markus Gasteiger
 * @date January, 2018
 * @brief With boundaries from boundary.h, the axis class allows one to define axes. Each axis represents one dimensions and therefor has two boundaries.
 *
 */

#ifndef axis_h
#define axis_h

#include <string>
#include <ostream>
#include <cassert>
#include <cmath>

#include "debug.h"
#include "boundary.h"

/**
* @brief Possible discretizations of the axis are described with this.
*
*/
enum class discretization {
    NONE, /**< dummy placeholder */
    HOMOGENEOUS /**< default discretization */
};


std::string discr_name(discretization _b){
    switch(_b)
    {
        case discretization::HOMOGENEOUS   : return "equidistant discretization"; break;
        default: assert(false && "No valid discretization type given."); // default
    }
    return "ERROR";
}

template<typename T, typename I, char NAME, discretization D = discretization::HOMOGENEOUS, class ... f_args>
struct axis_hg;
/**
* @brief The axis class accumulates all the information that is necessary to describe a single axis of a multidimensional space. LOWER and UPPER should be deduced automatically, but are not. Hence, the factory function make_axis is necessary.
*
* \tparam T is the floating point type.
* \tparam I is the integer type.
* \tparam NAME is the single char name of that axis, e.g. 'x'.
* \tparam D represents the type of discretization.
* \tparam f_args...
*
*/
//\TODO: test
template<typename T, typename I, char NAME, discretization D = discretization::NONE, class ... f_args>
struct axis{
    /**
    * @brief Constructor to the axis.
    *
    * @param _min is the beginning value of the interval.
    * @param _max is the maximum value in that dimension.
    * @param _lower_boundary is the lower boundary.
    * @param _upper_boundary is the upper boundary.
    */
    axis(const T& _min, const T& _max, size_t P,
         const boundary::boundary<T,I,f_args...>& _lower_boundary = boundary::boundary<T,I,f_args...>(boundary::btype::NONE,0),
         const boundary::boundary<T,I,f_args...>& _upper_boundary = boundary::boundary<T,I,f_args...>(boundary::btype::NONE,0),
         const std::string& _name = std::string(1,NAME), discretization _D = D, bool _HD = discretization::HOMOGENEOUS == D)
    :   dscr(_D), homogeneous(_HD), name(_name),
        points(P), gc_lower(_lower_boundary.gc), gc_upper(_upper_boundary.gc), size(gc_lower+points+gc_upper),
        min(_min), max(_max), delta((_max-_min)/(P-1)),
        boundaries({_lower_boundary,_upper_boundary}) {
        assert(P>1 && "More than one grid point is neccesary.");
        assert(_max>_min);
    }

    /**
    * @brief The discretization type.
    *
    */
    constexpr static discretization cdscr = D;
    const discretization dscr;
    /**
    * @brief Is true if the discretization is homogeneous.
    *
    */
    constexpr static bool chomogeneous = discretization::HOMOGENEOUS == D;
    const bool homogeneous;
    /**
    * @brief The name of the given axis.
    *
    */
    constexpr static char cname = NAME;
    const std::string name;
    /**
    * @brief The number of points within the domain, i.e. without ghost cells.
    *
    */
    const size_t points;
    /**
    * @brief The number of lower ghost cells.
    *
    */
    const size_t gc_lower;
    /**
    * @brief The number of upper ghost cells.
    *
    */
    const size_t gc_upper;
    /**
    * @brief The total number of points, including ghost cells.
    *
    */
    const size_t size;

    /**
    * @brief The domain's interval.
    *
    */
    const T min, max;
    /**
    * @brief The step size within the domain.
    *
    */
    const T delta;

    virtual inline T internal_get_delta(I pos) const = 0;
    virtual inline T internal_get_distance(I _pos1, I _pos2) const = 0;
    virtual inline T internal_get_coord(I _pos) const = 0;
    virtual inline size_t internal_get_pos(T _value) const = 0;
    /**
    * @brief Get the step size at a specific position.
    *
    * @tparam offset is false, the _pos=0 at the lower limit, i.e. at the lower ghost cells.
    * @param _pos is the discrete position on the axis.
    * @return T
    */
    template<bool offset = true>
    inline T get_delta(I _pos) const;
    template<bool offset = true>
    inline T get_distance(const I& _pos1, const I& _pos2) const;
    /**
    * @brief Get the coordinate corresponding to the positions _pos on the axis.
    *
    * @param offset == false -> position null is after the lower ghost cells.
    * @param _pos the input position.
    * @return T
    */
    template<bool offset = true>
    inline T get_coord(I _pos) const;
    /**
    * @brief Get the position on the axis to a given coordinate value.
    *
    * @param offset if true, the position 0 is after the lower ghost cells.
    * @param _value is the value, to which the position must be computed.
    * @return I
    */
    template<bool offset = true>
    inline I get_pos(const T& _value) const;

    auto simplify() const {
        if constexpr (cdscr == discretization::HOMOGENEOUS) {
//             debug::msg2 << "homogeneous" << std::endl;
            return std::make_shared<axis_hg<T,I,' ', discretization::NONE, f_args...>>(axis_hg<T,I,' ', discretization::NONE, f_args...>(min, max, points, boundaries[0], boundaries[1], name, discretization::HOMOGENEOUS, true));
        } else if constexpr (cdscr == discretization::NONE) {
//             debug::msg2 << "none" << std::endl;
            return std::make_shared<axis<T,I,' ',discretization::NONE, f_args...>>(axis<T,I,' ',discretization::NONE, f_args...>(min, max, points, boundaries[0], boundaries[1], name, dscr, homogeneous));
        }//\todo: check for memory leaks
    }

    template<class ...new_args>
    auto shrink_stupid() const {
        static_assert(cdscr == discretization::NONE); // should only be used out of a simplified grid axis.
        assert(dscr != discretization::NONE);

        //NOTE: currently there is only one valid case possible. This code structure was chosen to avoid a warning and fix the output type.
        if (dscr != discretization::HOMOGENEOUS) {
            assert(false);
        }

        return std::make_shared<axis_hg<T,I,' ',discretization::NONE, new_args...>>(axis_hg<T,I,' ',discretization::NONE, new_args...>(min, max, points, boundaries[0].template shrink_stupid<new_args...>(), boundaries[1].template shrink_stupid<new_args...>(), name, discretization::HOMOGENEOUS, true));
    }

    /**
    * @brief The boundary objects.
    *
    */
    std::array<boundary::boundary<T,I,f_args...>,2> boundaries;
//     boundary::boundary<T,I,f_args...>& boundaries[1] = boundaries[1];

    std::string print() const {
        std::ostringstream out;

        out << "axis \'" << name << "\': ";
        out << "\t";
        if (boundaries[0].type != boundary::btype::NONE){
            out << boundaries[0].type << "| " << boundaries[0].gc << "(" << boundary::name(boundaries[0].type) << ") " <<  boundary::btype::NONE;
        }
        out << " |"<< name <<"_min=" << min << "   "  << points << "(" << discr_name(dscr) << ")" << "   "<< name <<"_max=" << max << "| ";
        if (boundaries[1].type != boundary::btype::NONE){
            out  << boundaries[1].type << boundaries[1].gc << "(" << boundary::name(boundaries[1].type) << ") " << " |" << boundary::btype::NONE;
        }

        return out.str();
    }

protected:
    /**
    * @brief This function checks, whether the given position is within the limits of the axis.
    *
    * @tparam offset observe the offset position after the lower ghost cells.
    * @param _pos is the position to check.
    */
    template<bool offset = true>
    inline void check_position(const I& _pos) const;
};


/**
* @brief Specialization of the axis class with homogeneous discretization.
*
*/
template<typename T, typename I, char NAME, discretization D, class ... f_args>
struct axis_hg : axis<T,I,NAME,D,f_args...>{
    using axis<T,I,NAME,D,f_args...>::axis;

    /**
     * @brief Get the step size at a specific position.
     *
     * @tparam offset is true, the _pos=0 at the lower limit, i.e. at the lower ghost cells.
     * @param _pos is the discrete position on the axis.
     * @return T
     */
    virtual inline T internal_get_delta(I _pos) const override;
    virtual inline T internal_get_distance(const I _pos1, const I _pos2) const override;
    /**
     * @brief Get the coordinate corresponding to the positions _pos on the axis.
     *
     * @param offset == true -> position null is after the lower ghost cells.
     * @param _pos the input position.
     * @return T
     */
    virtual inline T internal_get_coord(I _pos) const override;
    /**
     * @brief Get the position on the axis to a given coordinate value.
     *
     * @param offset if true, the position starts after the lower ghost cells.
     * @param _value is the value, to which the position must be computed.
     * @return I
     */
    virtual inline size_t internal_get_pos(const T _value) const override;
};

/**
* @brief Decide on the axis class depending on the template parameters.
*
* \tparam T is the floating point type.
* \tparam I is the integer type.
* \tparam NAME is the single char name of that axis, e.g. 'x'.
* \tparam P is the number of grid points in the domain.
* \tparam D represents the type of discretization.
* \tparam LOWER is the lower boundary type.
* \tparam UPPER is the upper boundary type.
*
*/
template<typename T, typename I, char NAME, discretization D, class ... f_args>
struct decide_axis;/*{
    using type = axis<T, I, NAME, D, f_args...>;
};*/
template<typename T, typename I, char NAME, class ... f_args>
struct decide_axis<T, I, NAME, discretization::HOMOGENEOUS, f_args...>{
    using type = axis_hg<T, I, NAME, discretization::HOMOGENEOUS,f_args...>;
};

/**
* @brief Factory function for axis. This is necessary, because otherwise the automatic boundary type deduction would fail.
*
* @tparam T is the floating point type.
* @tparam I is the signed integer type.
* @tparam NAME is the name of the axis.
* @tparam P is the number of points within the domain.
* @tparam D is the discretization.
* @tparam b1 is the lower boundary type.
* @tparam b2 is the upper boundary type.
* @param _min is the begin of the domain.
* @param _max is the end coord of the domain.
* @param lower is the lower boundary.
* @param upper is the upper boundary.
* @return auto
*/
template<typename T, typename I, char NAME, discretization D = discretization::HOMOGENEOUS, class ... f_args>
auto make_axis(T _min, T _max, size_t P,
               const boundary::boundary<T,I,f_args...>& _lower_boundary = boundary::boundary<T,I,f_args...>(boundary::btype::NONE,0),
               const boundary::boundary<T,I,f_args...>& _upper_boundary = boundary::boundary<T,I,f_args...>(boundary::btype::NONE,0)){

    return typename decide_axis<T,I,NAME,D,f_args...>::type(_min,_max,P,_lower_boundary,_upper_boundary);
}


#include "axis.hpp"
#endif
