/**
 * @file boundary.h
 * @Author Markus Gasteiger
 * @date January, 2018
 * @brief Boundary types and corresponding boundaries are defined.
 *
 */

#ifndef boundary_h
#define boundary_h

#include <functional>
#include "functional_types.h"
#include "helpers.h"

namespace boundary{

/**
* @brief The possible boundary types.
*
*/
enum class btype {
    NONE, /**< placeholder, which should not be used in real problems */
    OPEN, /**< open boundary with no restrictions */
    DIRICHLET, /**< Dirichlet boundary condition */
    NEUMANN, /**< Neumann boundary condition */
    PERIODIC, /**< periodic boundary condition */
    MIRROR, /**< mirror boundary condition */
    NEIGHBOR, /**< internal boundary that is used to exchange grid data with parallel instances */
};

std::ostream& operator<<(std::ostream& os, btype _b)
{
    switch(_b)
    {
        case btype::DIRICHLET   : os << "\033[1;33m"; break; // yellow
        case btype::NEUMANN: os << "\033[1;32m"; break; // green
        case btype::PERIODIC : os << "\033[1;31m";  break; // red
        case btype::MIRROR  : os << "\033[1;35m";   break; // magenta
        case btype::NEIGHBOR  : os << "\033[1;36m";   break; // cyan
        default: os << "\033[0m"; // default
    }
    return os;
}

std::string name(btype _b){
    switch(_b)
    {
        case btype::DIRICHLET   : return "Dirichlet"; break;
        case btype::NEUMANN: return "Neumann"; break;
        case btype::PERIODIC : return "Periodic";  break;
        case btype::MIRROR  : return "Mirror";   break;
        case btype::NEIGHBOR  : return "Neighbor";   break;
        case btype::OPEN  : return "Open";   break;
        case btype::NONE  : return "No";   break;
        default: assert(false && "No correct boundary type given."); // default
    }
    return "ERROR";
}

/**
* @brief The base boundary class has all members that any of the boundary types would need.
* \tparam T is the floating point type.
* \tparam I is the integer valued type.
* \tparam ghost_cells basically describes the thickness of the boundary.
* \tparam BT is the corresponding boundary type.
* \tparam f_args... are the possible arguments to function describing the boundary.
*
*/
template<typename T, typename I, class ... f_args>
struct boundary{
    /**
    * @brief The boundary thickness as defined by ghost_cells.
    *
    */
    const size_t gc;
    /**
    * @brief The boundary type as defined by BT.
    *
    */
    const btype type;
    /**
    * @brief The function describing the boundary in case of Neumann or Dirichlet boundary conditions.
    *
    */
    const std::function<T(const f_args&...)> function;

    /**
    * @brief The boundary constructor basically only takes the function as an input argument.
    *
    * @param _type is this boundaries type.
    * @param _gc is the number of ghost cell layers.
    * @param _function is a std::function to describes the boundaries. It is only necessary for Neumann and Dirichlet boundary conditions.
    */
    boundary(btype _type, size_t _gc, const std::function<T(const f_args&...)>& _function = [](const f_args&...){return (T)0;}) : gc(_gc), type(_type), function(_function) {}

    /**
    * @brief A constructor for const value boundary funcitons.
    *
    * @param _type is this boundaries type.
    * @param _gc is the boundary thickness.
    * @param _val is the constant value that is used for the respective boundary.
    */
    boundary(btype _type, size_t _gc, const T _val) : gc(_gc), type(_type), function([_val](const f_args&...) {return _val;}) {}

    /**
    * @brief The boundary class copy constructor.
    *
    * @param _in is the input boundary.
    */
    boundary(const boundary<T,I,f_args...>& _in) : gc(_in.gc), type(_in.type), function(_in.function) {}

    boundary<T,I,f_args...>& operator=(const boundary<T,I,f_args...>& _input){
        if (this == &_input){
            return *this;
        }

        new (this) boundary<T,I,f_args...>(_input);
        return *this;
    }

    template<class ... new_args>
    boundary<T,I,new_args...> shrink_stupid() const {
        return boundary<T,I,new_args...>(type,gc,[](const new_args& ...)->T {assert(false && "The boundary function has been stupified!"); return 0;});
    }

};

}


#endif
