/**
 * @file export.h
 * @Author Markus Gasteiger
 * @date September, 2017
 * @brief A regular interface to save grid data.
 *
 */

//NOTE: useful example: https://www.unidata.ucar.edu/software/netcdf/examples/programs/PresTemp4dWr.cpp

#ifndef export_h
#define export_h

#include <string>
#include <memory>
#include <netcdf>
#include <string>
#include <array>
#include <vector>
#include <algorithm>
#include <type_traits>

#include "grid.h"
#include "properties.h"
#include "debug.h"


namespace EXPORT{

using namespace netCDF;
// using namespace netCDF::exceptions;

template<typename T, typename I, class NAMES, bool HD>
struct dummy {
public:
    dummy(const grid<T, I, NAMES, HD>& , const std::string& out_path) : path(out_path){};

    void write_initial_data(const grid<T, I, NAMES, HD>& _grid);
//     virtual void append_grid() = 0;
    virtual void write_grid(grid<T,I,NAMES,HD>& _grid) = 0;
    const std::string path;
};

template<typename T, bool is_fundamental = std::is_fundamental<T>::value, bool is_signed = std::is_signed<T>::value, bool is_int = std::is_integral<T>::value, bool is_float = std::is_floating_point<T>::value, size_t size = sizeof(T)> struct c_to_netcdf;

template<> struct c_to_netcdf<char, std::is_fundamental<char>::value, std::is_signed<char>::value, std::is_integral<char>::value, std::is_floating_point<char>::value, 1>{ using type = netCDF::NcChar;};
template<size_t S> struct c_to_netcdf<std::string, std::is_fundamental<std::string>::value, std::is_signed<std::string>::value, std::is_integral<std::string>::value, std::is_floating_point<std::string>::value, S>{ using type = netCDF::NcString;};

//NOTE: type size according to ncTypes.h It would be better, if the sizes of the underlying types were to be evaluated at compile time.
template<typename T> struct c_to_netcdf<T, true, true, true, false, 1>{ using type = netCDF::NcByte;}; // 8bit int
template<typename T> struct c_to_netcdf<T, true, false, true, false, 1>{ using type = netCDF::NcUbyte;}; // 8bit unsigned int
template<typename T> struct c_to_netcdf<T, true, true, true, false, 2>{ using type = netCDF::NcShort;}; // 16 bit int
template<typename T> struct c_to_netcdf<T, true, false, true, false, 2>{ using type = netCDF::NcUshort;}; // 16 bit unsigned int
template<typename T> struct c_to_netcdf<T, true, true, true, false, 4>{ using type = netCDF::NcInt;}; // 16 bit int
template<typename T> struct c_to_netcdf<T, true, false, true, false, 4>{ using type = netCDF::NcUint;}; // 16 bit unsigned int
template<typename T> struct c_to_netcdf<T, true, true, true, false, 8>{ using type = netCDF::NcInt64;}; // 16 bit int
template<typename T> struct c_to_netcdf<T, true, false, true, false, 8>{ using type = netCDF::NcUint64;}; // 16 bit unsigned int
template<typename T> struct c_to_netcdf<T, true, true, false, true, 4>{ using type = netCDF::NcFloat;}; // 16 bit int
template<typename T> struct c_to_netcdf<T, true, true, false, true, 8>{ using type = netCDF::NcDouble;}; // 16 bit int


template<typename T, typename I, class NAMES, bool HD, bool use_records=false>
struct netcdf : public dummy<T,I,NAMES,HD> {
public:
//     using namespace netCDF;

    netcdf(const grid<T, I, NAMES, HD>& _grid, const std::string& out_path) : dummy<T,I,NAMES,HD>(_grid, out_path), props(_grid.props), file(NcFile(out_path,NcFile::replace)){
//         assert(file.is_valid()); //NOTE: netCDF::is_valid() does not work?

        write_initial_data(_grid);
    }
    ~netcdf(){};

    void write_initial_data(const grid<T, I, NAMES, HD>& _grid) {
        std::vector<T> v;
        axes.resize(NAMES::size+use_records);
        axes_vars.resize(NAMES::size+use_records);
        if constexpr (use_records){
            axes[0] = file.addDim("time");
//             names[0] = "time";
        }

        size_t j=use_records+NAMES::size-1;
        for(size_t i=0; i<NAMES::size; ++i, --j){
            names[j] = std::string(1,NAMES::list[i]);
            axes[j] = file.addDim(names[j], _grid.props.points[i]);

            axes_vars[j] = file.addVar(names[j], typename c_to_netcdf<T>::type(), axes[j]);
            v.resize(_grid.props.points[i]);
            for(size_t k=0; k<_grid.props.points[i]; ++k){
                v[k]=_grid.axes[i]->template get_coord<true>(k);
            }
            axes_vars[j].putVar(v.data());
        }
        data = file.addVar("f", typename c_to_netcdf<T>::type(), axes);
    }

    virtual void write_grid(grid<T,I,NAMES,HD>& _grid) override {
        std::vector<size_t> startp(NAMES::size+use_records);
        std::vector<size_t> countp(NAMES::size+use_records);
        std::vector<size_t> stridep(NAMES::size);
        std::array<size_t,NAMES::size> step;
        if constexpr (use_records) {
            startp[0]=_grid.age;
        }
        startp.back() = 0;
        std::fill(stridep.begin(), stridep.end(), 1);
        step.fill(0);
        step[1]=1;

        std::fill(countp.begin(), countp.end(), 1);
        countp.back() = this->props.points[0];

        auto itr = _grid.template begin_intern<false,false>(this->props.limit_lower, this->props.limit_upper);

        size_t run =0;
        debug::msg3 << _grid.props.limit_lower << " - " << _grid.props.limit_upper << std::endl;

        for(; itr.within(); itr+=step){
            for(size_t k=0; k<NAMES::size-1; ++k){
                startp[k+use_records]=itr.crds[NAMES::size-1-k]-itr.begin[NAMES::size-1-k];
            }

            debug::msg3 << ++run << ": " << itr.crds << " - " << startp << std::endl;

            data.putVar(startp, countp, itr.ptr);
        }
    }

    const properties<T,I,NAMES> props;

private:
    const NcFile file;
    NcVar data;
    std::vector<NcDim> axes;
    std::vector<NcVar> axes_vars;
    std::array<std::string, NAMES::size+use_records> names {};
};

}
#endif
