/**
 * @file grid.h
 * @Author Markus Gasteiger
 * @date January, 2017
 * @brief The grid class combines dim independent axes, which are defined in axis.h, and combines them to a dim dimensional space.
 *
 */

#ifndef grid_h
#define grid_h

#include <memory>
#include <ostream>
#include <string>
#include <array>
#include <type_traits>

#include "axis.h"
#include "liste.h"
#include "helpers.h"
#include "dim.h"
#include "debug.h"
#include "optimize.h"
#include "boundary.h"
#include "reduction.h"
#include "properties.h"
#include "grid_iterator_delta_coords.h"


template<typename T, typename I, class NAMES = liste<char>, bool HD = false>
struct grid : std::enable_shared_from_this<grid<T,I,NAMES,HD>> {
    // own type
    using this_type = grid<T,I,NAMES,HD>;
    using dims = NAMES;
    constexpr static bool is_HD = HD;

//     constructors
    grid() : props(), limits({}), axes({}), storage(0) { age=0;}

    grid(const properties<T,I,NAMES>& _props, const std::array<std::shared_ptr<axis<T,I,' ',discretization::NONE,std::array<T,NAMES::size>>>,NAMES::size>& _axes) : props(_props), limits(set_limits()), axes(_axes), storage(props.size) {
        age=0;
        if constexpr (HD){
            deltas = get_deltas<false>(std::array<size_t,NAMES::size> {});
        }
    }


//     template<typename ...AXES>
//     grid(const AXES& ... _axes) : props(_axes...), limits(set_limits()), axes({(std::static_pointer_cast<axis<T, I, ' ', discretization::NONE,std::array<T,NAMES::size>> >(_axes.simplify()))...}), storage(props.size) {
//         age=0;
//         if constexpr (HD){
//             deltas = get_deltas<false>(std::array<size_t,NAMES::size> {});
//         }
//     }


    template<bool HD_new>
    grid(const grid<T,I,NAMES,HD_new>& input) : props(input.props), limits(input.limits), axes(input.axes), storage(props.size) {
        static_assert(HD||(!HD_new),"A homogeneously discretized grid cannot be made non-homogeneous.");
        age=input.age;
        deltas = input.deltas;
    }

    ~grid(){
        for(auto iter = axes.begin(); iter!= axes.end(); ++iter){
            iter->reset();
        }
    }

    // custom copy assignment operator
    this_type& operator=(const this_type& input){
        if(this == &input){
            return *this;
        }
        new (this) this_type(input);
        return *this;
    }

    // general stuff
    const properties<T,I,NAMES> props;
    constexpr static size_t dim = NAMES::size;
    std::array<T,dim> deltas;
    inline size_t size(){ return props.size;}

    const std::array<std::array<std::array<size_t,dim>,4>,dim> limits;

    // data
    size_t age;

    std::array<std::shared_ptr<axis<T,I,' ',discretization::NONE,std::array<T,dim>>>,dim> axes;
    void change_boundaries(const char _b, const bool is_upper, const boundary::boundary<T,I,std::array<T,dim>>& _new);
    std::vector<T> storage; // is local ... should only be copied deep.

    // functions
    template<bool offset=true, typename ITER = typename std::vector<T>::iterator>
    inline ITER coords2pos(const std::array<I,dim>&) const;

    template<bool offset=true, typename ITER = typename std::vector<T>::iterator>
    inline std::array<I,NAMES::size> pos2coords(ITER) const;

    template<bool offset=true, typename I2>
    inline std::array<T,NAMES::size> get_deltas(const std::array<I2,dim>&) const ;
    template<bool offset=true, typename I2>
    inline std::array<T,NAMES::size> get_coords(const std::array<I2,dim>&) const ;
    template<bool offset=true, typename I2>
    inline T& operator[](I2) ;
    template<bool offset=true, typename I2>
    inline T& operator[](const std::array<I2,dim>&);
    inline T& operator()(const std::array<T,dim>&);

    // iterator
    /**
    * @brief The nest iterator gives an interface to loop through the contents within a dimensional specific range. To do so, it is necessary to keep a pointer to the encapsulating grid class. If it is used intrinsically, a raw pointer is sufficient to achieve that. But if the iterator were to be used externally, it is necessary to use a shared_ptr to satisfy the objects existence. But when wants to use a shared_ptr of this a controlling instance of shared_ptr has already to exist outside to secure the existence. Hence, there is a choice of POINTER that allows one to to exactly that: internally one can just enter 'this', but externally it is necessary to call 'shared_from_this'. But all this should be hidden away within 'begin'.
    *
    */
    template<typename POINTER = std::shared_ptr<grid<T,I,NAMES,HD>>, template <typename,bool,typename,typename,class> class = basic_iterator>
    struct iterator;

    template<template <typename,bool,typename,typename,class> class = basic_iterator>
    inline auto begin(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end) ;
    template<bool use_deltas = false, bool use_coords = false>
    inline auto begin(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end){
        return begin<iter_type<use_deltas,use_coords>::delta_coords>(begin,end);
    }
//     template<bool use_deltas = false, bool use_coords =  false>
//     inline auto begin(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end) ;
    template<template <typename,bool,typename,typename,class> class = basic_iterator>
    inline auto begin() { return begin<iter_type>(props.limit_lower, props.limit_upper);}
    template<bool use_deltas = false, bool use_coords = false>
    inline auto begin() { return begin<iter_type<use_deltas,use_coords>::template delta_coords>(props.limit_lower, props.limit_upper);}

    inline auto end(const std::array<size_t,dim>& end) const;


//// grid_functions.h
    template<template <typename,typename> class R = reduction::SUM,
            class new_names,
            typename FUNCTION,
            bool iter_direct,
            template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    auto reduction(const FUNCTION& _f);
    template<template <typename,typename> class R = reduction::SUM,
            class new_names,
            bool use_deltas, bool use_coords,
            typename FUNCTION>
    auto reduction(const FUNCTION& _f);
    template<template <typename,typename> class R = reduction::SUM,
            class new_names,
            template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
            auto reduction(const std::function<T(const iterator<this_type*,ITER_TYPE>&)>& _f);

    template<class REDUCTION = reduction::SUM<T,I>,
             typename FUNCTION = typename std::function<T(const std::array<T,dim>&)>,
             bool iter_direct,
             template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    void reduction_function(REDUCTION& R, const FUNCTION& _f, iterator<this_type*,ITER_TYPE> range);

    template<typename FUNCTION = typename std::function<T(const std::array<T,dim>&)>,
             bool access=false,
             typename POINTER = std::shared_ptr<grid<T,I,NAMES,HD>>,
             bool iter_direct=false,
             template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    void function(const FUNCTION&, iterator<POINTER,ITER_TYPE> range);

    template<typename FUNCTION = typename std::function<T(const std::array<T,dim>&)>,
            bool access=false,
            typename POINTER = std::shared_ptr<grid<T,I,NAMES,HD>>,
            bool iter_direct=false,
            template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    void function(const FUNCTION&, iterator<POINTER,ITER_TYPE> range, grid<T,I,NAMES,HD>& output);

    template<bool access=false,
            template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    void function(const typename function_type_e<NAMES::size,T,I,T*,access,ITER_TYPE<this_type*,HD,T,I,NAMES>>::type&);
    template<class VALUE>
    void function(const std::function<T(const VALUE& )>& );

    template<bool access, bool use_deltas, bool use_coords>
    void function(const typename function_type<NAMES::size,T,I,T*,access,use_deltas,use_coords>::type& _f);
    template<template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    void function(const std::function<T(const iterator<this_type*,ITER_TYPE>&)>& _f);

    void update_boundaries();

    T* data() { return storage.data();}
    grid<T,I,NAMES,HD>* THIS(){return this;}

    template<template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    inline auto begin_intern(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end){
            return grid<T,I,NAMES,HD>::iterator<this_type*, ITER_TYPE> (this,begin,end);
        }
    template<template <typename,bool,typename,typename,class> class ITER_TYPE = basic_iterator>
    inline auto begin_intern() {
        return begin_intern<ITER_TYPE>(props.limit_lower, props.limit_upper);
    }
    template<bool use_deltas = false, bool use_coords = false>
    inline auto begin_intern(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end){
        return begin_intern<iter_type<use_deltas,use_coords>::template delta_coords>(begin, end);
    }

    std::string print( const iterator<this_type*,basic_iterator>& range, const std::function<T(T)> _function = [](T _in)->T {return _in;} );
    std::string print() { return print(begin_intern<basic_iterator>());}
    std::string print(const std::array<size_t,dim>& _begin, const std::array<size_t,dim>& _end, const std::function<T(T)> _function = [](T _in)->T {return _in;} ) { return print(begin_intern<basic_iterator>(_begin, _end), _function);}
    std::string print_all(const std::function<T(T)> _function = [](T _in)->T {return _in;} ) { return print(begin_intern<basic_iterator>(std::array<size_t,dim> {},props.sizes), _function);}

    std::string data_info();


    template<class sub_names>
    grid<T,I,sub_names,HD> sub_grid(){
        using ar_type = std::array<std::shared_ptr<axis<T,I,' ',discretization::NONE,std::array<T,sub_names::size>>>,sub_names::size>;
        ar_type new_ax {};
        std::array<size_t,sub_names::size> sub_pos = NAMES::template sub_dim_pos<sub_names>();
        properties<T,I,sub_names> new_props = props.template sub_space<sub_names>();
        for(size_t i=0; i<sub_names::size; ++i){
            new_ax[i]=axes[sub_pos[i]]->template shrink_stupid<std::array<T,sub_names::size>>();
        }

        return grid<T,I,sub_names,HD>(new_props,new_ax);
    }
protected:
    auto set_limits();
    void copy_internal(const iterator<this_type*,basic_iterator>& source, const iterator<this_type*,basic_iterator>& target);
    void copy_shift(iterator<this_type*,basic_iterator> target, std::make_signed<size_t>::type shift);

    void derivative_on_range(size_t axis, bool is_upper, const std::array<size_t,dim>& _begin, const std::array<size_t,dim>& _end, const std::function<T(const std::array<T,dim>&)>&);

    template<bool access=false, bool use_deltas=false, bool use_coords=true>
    void function_on_inner(const typename function_type<NAMES::size,T,I,T*,access,use_deltas,use_coords>::type&, const grid<T,I,NAMES,HD>& _out);

};


//TODO: replace with template deduction guide.
template<typename T, typename I, bool SHARED = false, class ... AXES>
auto make_grid(const AXES& ... _axes){
    using NAMES_new = liste<char, AXES::cname ...>;
    static_assert(NAMES_new::is_uniq(),"Axes have to be independent.");
    const bool HD_new = (... && AXES::chomogeneous);

    properties<T,I,NAMES_new> new_props(_axes...);

    const std::array<std::shared_ptr<axis<T,I,' ',discretization::NONE,std::array<T,NAMES_new::size>>>,NAMES_new::size>& axes {(std::static_pointer_cast<axis<T, I, ' ', discretization::NONE,std::array<T,NAMES_new::size>> >(_axes.simplify()))...};

    if constexpr (SHARED) {
        return std::make_shared<grid<T,I,NAMES_new, HD_new>> (new_props, axes);
    } else {
        return grid<T,I,NAMES_new, HD_new> (new_props, axes);
    }
}


#include "grid.hpp"
#include "grid_functions.hpp"
#include "grid_iterator.h"
#include "grid_helpers.h"

#endif
