#ifndef properties_h
#define properties_h

// #include "grid.h"
#include "axis.h"
#include "helpers.h"

template<typename T, typename I, class NAMES>
struct properties{
    properties(bool _homogeneous,
               const std::array<size_t,NAMES::size>& _gc_lower,
               const std::array<size_t,NAMES::size>& _points,
               const std::array<size_t,NAMES::size>& _gc_upper) : hd(_homogeneous),
    gc_lower(_gc_lower), points(_points), gc_upper(_gc_upper), limit_lower(gc_lower),
    limit_upper(_gc_lower+_points),
    sizes(limit_upper+_gc_upper),
    strides(make_strides(sizes)),
    size(get_size(sizes)),
    hash(hash_array<NAMES::size,T,size_t>(strides)){
        initialize();
    }

    template<class ... AXES>
    properties(const AXES& ... _axes) : properties((...&&_axes.homogeneous), {_axes.gc_lower...}, {_axes.points...}, {_axes.gc_upper...}) {}

    properties(const properties<T,I,NAMES>& _in) : hd(_in.hd), gc_lower(_in.gc_lower), points(_in.points), gc_upper(_in.gc_upper), limit_lower(_in.limit_lower), limit_upper(_in.limit_upper), sizes(_in.sizes), strides(_in.strides), size(_in.size), hash(_in.hash) {
        initialize();
    }

    properties() : properties(false, std::array<size_t,NAMES::size> {}, emplace_array<size_t,NAMES::size>(1), std::array<size_t,NAMES::size> {}) {};

    properties<T,I,NAMES>& operator=(const properties<T,I,NAMES>& input){
        if(this == *input){
            return *this;
        }
        new (this) properties<T,I,NAMES>(input);
        return *this;
    }

    const bool hd;
    constexpr static size_t dim = NAMES::size;
    constexpr static std::array<size_t,dim> zeroes {};
    const std::array<size_t,dim> gc_lower;
    const std::array<size_t,dim> points;
    const std::array<size_t,dim> gc_upper;
    const std::array<size_t,dim> limit_lower;
    const std::array<size_t,dim> limit_upper;
    size_t linear_lower_limit;
    size_t linear_upper_limit;
    const std::array<size_t,dim> sizes;
    const std::array<size_t,dim> strides;
    const size_t size;
    const size_t hash;

    template<bool offset=true, typename IN>
    inline IN inner_product(const std::array<IN,dim>& _in) const {
        IN out = 0;
        auto ss = strides.cbegin();
        auto is = _in.cbegin();
        auto ls = limit_lower.cbegin();
        for(; ss!=strides.cend(); ++is, ++ss, ++ls){
            if constexpr (offset) {
                out += *ss * (*is + *ls);
            } else {
                out += *ss * *is;
            }
        }

        return out;
    }

    template<bool offset=true, typename IN>
    inline std::array<IN,dim> pos2coords(IN _in) const {
        std::array<IN,dim> output {};
        auto os = output.rbegin();
        auto ss = strides.crbegin();
        auto ls = limit_lower.crbegin();
        for(; ss!=strides.crend(); ++os, ++ss, ++ls){
            *os = (IN) (_in/ *ss);
            _in = _in % *ss;
            if constexpr (offset) { *os -= *ls;}
        }

        return output;
    }

    template<bool offset=true, typename IN>
    bool within_domain(const std::array<IN,NAMES::size>& _in) const {
        auto iter_l = limit_lower.cbegin();
        auto iter_u = limit_upper.cbegin();
        if constexpr (offset) {
            iter_l = zeroes.cbegin();
            iter_u = sizes.cbegin();
        }
        auto iter = _in.cbegin();
        for(; iter != _in.cend(); ++iter_l, ++iter, ++iter_u){
            if ((I)*iter_l < -*iter || (I)*iter_u< *iter) {return false;};
        }

        return true;
    }

    template<class sub_names>
    properties<T,I,sub_names> sub_space() const {
        static_assert(sub_names::is_uniq());

        std::array<size_t,sub_names::size> gc_lower_new {};
        std::array<size_t,sub_names::size> points_new {};
        std::array<size_t,sub_names::size> gc_upper_new {};

        std::array<size_t,sub_names::size> sub_pos = NAMES::template sub_dim_pos<sub_names>();

        for(size_t i=0; i<sub_names::size; ++i){
            gc_lower_new[i]=gc_lower[sub_pos[i]];
            points_new[i]=points[sub_pos[i]];
            gc_upper_new[i]=gc_upper[sub_pos[i]];
        }

        return properties<T,I,sub_names>(hd,
                          gc_lower_new,
                          points_new,
                          gc_upper_new);
    }

protected:
    constexpr static std::array<size_t,dim> make_strides(const std::array<size_t,dim>& _in) {
        std::array<size_t,dim> out {};
        out.front()=1;
        auto iout = out.begin()+1;
        for(auto iin = _in.cbegin(); iout!=out.cend(); ++iout, ++iin) {
            *iout = *std::prev(iout) * *iin;
        }
    return out;
    }

    constexpr static size_t get_size(const std::array<size_t,dim>& _in) {
        size_t out = 1;
        for(auto& el: _in){
            out *= el;
        }
        return out;
    }

    void initialize(){
        linear_lower_limit = inner_product<false>(limit_lower);
        linear_upper_limit = inner_product<false>(limit_upper);
    }
};


template<typename T, typename I, class NAMES>
bool operator==(const properties<T,I,NAMES>& _first, const properties<T,I,NAMES>& _second){
    if ( _first.hd != _second.hd ) { return false;}
//     if ( _first.dim != _second.dim ) { return false;}
    if ( _first.gc_lower != _second.gc_lower ) { return false;}
//     if ( _first.points != _second.points ) { return false;}
    if ( _first.gc_upper != _second.gc_upper ) { return false;}
//     if ( _first.limit_lower != _second.limit_lower ) { return false;}
//     if ( _first.limit_upper != _second.limit_upper ) { return false;}
//     if ( _first.sizes != _second.sizes ) { return false;}
    if ( _first.strides != _second.strides ) { return false;}
    if ( _first.size != _second.size ) { return false;}
    if ( _first.hash != _second.hash ) { return false;}

    return true;
}


#endif
