#ifndef GMRES_h
#define GMRES_h

#include <limits>
#include <vector>

#include "norms.h"
#include "grid.h"
#include "stencil_func.h"
#include "optimize.h"
#include "debug.h"

template<typename T, typename I, typename STENCIL, typename GRID, template <typename> class NORM = L2>
struct arnoldi{
    // NOTE: memory reserving may not be very good.
    arnoldi(const STENCIL& _stencil, const GRID& _x0) : stencil(_stencil), Q({_x0}) {
        T tmp = norm<T,NORM>(Q[0]);
        Q[0].function<true,false,false>([tmp](T* _in)->T{return *_in / tmp;});
    }
    arnoldi(const STENCIL& _stencil, size_t size) : stencil(_stencil) {
        Q.reserve(size);
    }

    const STENCIL stencil;
    std::vector<GRID> Q;
    std::vector<std::vector<T>> hessenbergQR;

    void reset(const GRID& _x0){
        hessenbergQR.resize(0);
        Q.resize(1);
        Q[0]=_x0;
    }
    T reset(GRID& _x0, const GRID& _b){
        T output = 0;
        hessenbergQR.resize(0);
        Q.resize(1);

        Q[0]= stencil*_x0;

        auto iter_b = _b.storage.cbegin();
        for(auto iter = Q[0].storage.begin(); iter!=Q[0].storage.end(); ++iter, ++iter_b){
            *iter = *iter_b - *iter;
            output+=norm<T,NORM>::direct(*iter);
        }
        output = norm<T,NORM>::back(output);

        for(auto& el: Q[0].storage){
            el/=output;
        };

        return output;
    }

    bool add_dim(){
        std::vector<T> h_a_new;
        h_a_new.reserve(Q.size()+1);
        std::vector<size_t> h(0);

        Q.push_back(stencil*Q.back());

        auto new_iter = Q.back().storage.begin();
        for(auto iter = Q.cbegin(); iter!=Q.cend()-1; ++iter){
            h_a_new.push_back(norm<T,NORM>::direct(*iter,Q.back()));

            new_iter = Q.back().storage.begin();
            for(auto old_iter = iter->storage.cbegin(); old_iter!=iter->storage.cend(); ++old_iter, ++new_iter){
                *new_iter-=h_a_new.back()* *old_iter;
            }
        }
        h_a_new.push_back(norm<T,NORM>::self(Q.back()));
        if(is_about_zero(h_a_new.back())){
            return true;
        }

        for(new_iter = Q.back().storage.begin(); new_iter!=Q.back().storage.end(); ++new_iter) {
            *new_iter/=h_a_new.back();
        }

        hessenbergQR.push_back(h_a_new);

        return false;
    }

    void update_x(std::vector<T>& x, const std::vector<T>& y) const {
        auto iter = x.begin();
        auto iter_y = y.cbegin();
        for(auto iter_q=Q.cbegin(); iter_q!=Q.cend()-1; ++iter_q, ++iter_y){
            assert(x.size() == iter_q->storage.size());
            iter = x.begin();
            for(auto iter_q_data=iter_q->storage.cbegin(); iter_q_data!=iter_q->storage.cend(); ++iter, ++iter_q_data){
                *iter+=*iter_y * *iter_q_data;
            }
        }
    }

    std::vector<T> get_y(std::vector<T> output) const {
        assert(output.size() <= Q.size());
        for (std::make_signed<size_t>::type k = output.size()-2; k>=0; --k){
            assert(!is_about_zero(hessenbergQR[k][k]));
            output[k]/=hessenbergQR[k][k];
            for(std::make_signed<size_t>::type j=k-1; j>=0; --j){
                output[j]-=hessenbergQR[k][j]*output[k];
            }
        }

        return output;
    }

};


// saad 6.5.3
template<typename T>
void givens_rotation(const T v1, const T v2, T& cs, T& sn){
    if(is_about_zero(v1)){
        cs=0.;
        sn=1.;
    } else {
        cs = std::abs(v1)/std::sqrt(v1*v1+v2*v2);
        sn = cs*v2/v1;
    }
}

template<typename T>
void givens_transform(std::vector<T>& h, std::vector<T>& cs, std::vector<T>& sn){
    T tmp = 0;
    //     auto iter_h = h.begin();
    auto iter_cs = cs.cbegin();
    auto iter_sn = sn.cbegin();
    auto iter_h = h.begin();
    for(;iter_cs!=cs.cend(); ++iter_h, ++iter_cs, ++ iter_sn){
        tmp = *iter_cs * *iter_h + *iter_sn * *(iter_h+1);
        *(iter_h+1) = -*iter_sn * *iter_h + *iter_cs * *(iter_h+1);
        *iter_h = tmp;
    }
    cs.push_back(0.);
    sn.push_back(0.);
    givens_rotation(*iter_h, *(iter_h+1), cs.back(), sn.back());
    *iter_h = *iter_cs * *iter_h + *iter_sn * *(iter_h+1);
    *(iter_h+1) = 0.;
}

template<typename T, typename I, class NAMES, bool HD, bool is_function, bool is_deltas, bool is_coords,template <typename> class NORM = L2, class ... f_args>
struct gmres{
    using GRID = grid<T,I,NAMES,HD>;
    using STENCIL = stencil_func<T,I,NAMES,HD,is_function,is_deltas,is_coords,f_args...>;

    gmres(const STENCIL& _stencil,
          const GRID& _b,
          const size_t& _max_iterations = 1e3,
          const size_t& _restart = 20,
          const T& _threshold = 100*std::numeric_limits<T>::epsilon()) : stencil(_stencil), b(_b), b_norm(norm<T,NORM>::self(_b)), threshold(_threshold), max_iterations(_max_iterations), restart(_restart) {
              if(max_iterations < restart){
                  debug::msg2 << "Warning: the maximum number of iterations is smaller than the restart value." << std::endl;
              }
          }

    size_t iterations=0;
    const STENCIL& stencil;
    const GRID& b;
    const T b_norm;
    const T threshold;
    const size_t max_iterations;
    const size_t restart;

    size_t operator()(GRID& x){
        arnoldi<T,I,STENCIL,GRID,NORM> krylov(stencil,restart);

        std::vector<T> residual(1); residual.reserve(restart);
        std::vector<T> error; error.reserve(restart);
        std::vector<T> y; y.reserve(restart);
        std::vector<T> sn; sn.reserve(restart);
        std::vector<T> cs; cs.reserve(restart);

        for(iterations=0; iterations<max_iterations;){
            residual.resize(1);
            error.resize(1);
            residual[0] = krylov.reset(x,b);
            error[0]  = residual[0]/b_norm;
            if (error.back()<threshold) {
                return iterations;
            }
            sn.resize(0);
            cs.resize(0);

            for(size_t j=0; j<restart; ++j, ++iterations){
//                 debug::msg2 << "\r iterations: "<< iterations ;
                if(krylov.add_dim()){
                    y=krylov.get_y(residual);
                    krylov.update_x(x.storage, y);
                    return iterations;
                }

                givens_transform(krylov.hessenbergQR.back(), cs, sn);
                residual.push_back(-sn.back()*residual.back());
                *(residual.end()-2)*=cs.back();

                error.push_back(std::abs(residual.back())/b_norm);
                if (error.back() < threshold) {
                    break;
                }
            }

            y=krylov.get_y(residual);
            krylov.update_x(x.storage, y);
            debug::output << "iteration " << iterations << " with error "<< error.back()<<"/"<< threshold;
            if (error.back() < threshold) {
                debug::output << "-> break" << std::endl;
                break;
            }
            debug::output << std::endl;
        }

        debug::msg2 << "GMRES reached max_iterations with error " << error.back()<<"/"<< threshold << std::endl;
        return iterations;
    }

    GRID operator()(){
        GRID output = b;
        operator()(output);
        return output;
    }
};


#endif
