/**
 * @file more_complex_test.cpp
 * @Author Markus Gasteiger
 * @date August, 2018
 * @brief In this example we look at a little more complicated 1+1 system which consists of electrons and ions. For the 1+1 case the electron-ion collision term becomes $$ C_{ei} = -\Gamma n_i \left( \frac{1}{2} \left( \frac{1}{|v|} -\frac{sign(v)}{v^2} \right)\partial_v f \right) + \left(-\frac{sign(v)}{v^2} +2\frac{sign(v)}{v^3} \right) f .$$
 *
 * Hence, the equation looks like
 * $$ v \partial_z f+\frac{e E_z}{m} \partial_v f = C_{ei} $$
 *
 */
#include <cmath>
#include <vector>

#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"
#include "stencil_func.h"
#include "gmres.h"
#include "constants.h"
#include "boundary.h"
#include "helpers.h"
#include "export.h"
#include "helpers.h"

using T = double;
using I = long;


T T_i = 3000;
T T_e = 3000;
T m_i = 4*amu;

T n_i_0 = 1e18;
T n_e_0 = 1e18;

T c_s = v_thermal_K(T_i+T_e, m_i);
T c_e = v_thermal_K(T_e, m_e);
T c_i = v_thermal_K(T_i, m_i);

T Gamma_i = std::pow(q_e,4)*landau/4/M_PI/std::pow(epsilon_0*m_i,2);

inline T n_i(T v){ return n_i_0*maxwellian(v,c_i,c_s);}
inline T n_e(T v){ return n_e_0*maxwellian(v,c_e);}

// boundary::boundary<T,I,std::array<T,2>> HN(boundary::btype::NEUMANN,1,0.);
// boundary::boundary<T,I,std::array<T,2>> HGb(boundary::btype::DIRICHLET,1,[](const std::array<T,2>& c)->T{return std::exp(-c[1]*c[1]);});
//
// auto x_axis = make_axis<T,I,'x'>(0.,1.,1000,HGb,HGb);
// auto v_axis = make_axis<T,I,'v'>(-5*c_e,5*c_e,1000,HN,HN);
// auto grid = make_grid<T,I,false>(x_axis,v_axis);



boundary::boundary<T,I,std::array<T,2>> HN(boundary::btype::NEUMANN,1,0.);
boundary::boundary<T,I,std::array<T,2>> HD(boundary::btype::DIRICHLET,1,0);
boundary::boundary<T,I,std::array<T,2>> HGe(boundary::btype::DIRICHLET,1,[](const auto& c)->T{return n_e(c[1]);});
boundary::boundary<T,I,std::array<T,2>> HGi(boundary::btype::DIRICHLET,1,[](const auto& c)->T{return n_i(c[1]);});
boundary::boundary<T,I,std::array<T,2>> P(boundary::btype::PERIODIC,1);
boundary::boundary<T,I,std::array<T,2>> M2(boundary::btype::MIRROR,1);
boundary::boundary<T,I,std::array<T,1>> M1(boundary::btype::MIRROR,1);

auto ixspace = make_axis<T,I,'x'>(0.,.01,12,HGi,M2);
auto exspace = make_axis<T,I,'x'>(0.,.002,500,HGe,HGi);
auto vspace = make_axis<T,I,'v'>(-5*c_e,5.*c_e,501,HN,HN);
auto qxspace = make_axis<T,I,'x'>(0.,.01,12,M1,M1);

auto e_space = make_grid<T,I,false>(exspace,vspace);
auto i_space = make_grid<T,I,false>(ixspace,vspace);
auto q_space = make_grid<T,I,false>(qxspace);

auto dx_p = stencil<T,I,dim<'x'>>(epsilon_0,{-1.,1},{{1},{0}}).get_delta_f([](T val, const auto& d)->T {return val/d[0];});

int main() {
    // initialize the grids
    e_space.template function<false,true,true>([](const auto& d, const auto& c)->T{return n_e(c[1]);});
    i_space.template function<false,true,true>([](const auto& d, const auto& c)->T{return (1-c[0]/1e14)*n_i(c[1]);});

    //
    auto ne=e_space.template reduction<reduction::SUM,dim<'x'>,basic_iterator>([](const auto& c)->T {return  *c * c.deltas[1];});
    auto ni=i_space.template reduction<reduction::SUM,dim<'x'>,basic_iterator>([](const auto& c)->T {return  *c * c.deltas[1];});

    for(size_t i=0; i<q_space.storage.size(); ++i){
        q_space.storage[i] = -q_e*ne.storage[i]+q_e*ni.storage[i];
    }
    q_space.update_boundaries();
    auto dx = dx_p.adjust_to_grid(q_space);
    auto E=gmres(dx,q_space)();



//vlasov proto
// auto x_adv = stencil<T,I,dim<'x','v'>>(.5, {-1.,1},{{-1,0},{1,0}}).get_delta_coords_f([](T val, const auto& d, const auto& c)->T{return val/d[0]*c[1];},"momentum");
// centered
// auto v_adv_e = stencil<T,I,dim<'x','v'>>(-.5*q_e/m_e,{-1.,1},{{0,-1},{0,1}})
//                 .get_delta_coords_f([&E](T val, const auto& d, const auto& c)->T{
//                     return val*E(std::array<T,1>{c[0]});
//                 },"force_e"); // could be made much simpler with abstract iterator functions -> the x axis is the same -> the coordinate is the same;
// auto v_adv_i = stencil<T,I,dim<'x','v'>>(.5*q_e/m_i,{-1.,1},{{0,-1},{0,1}})
//                 .get_delta_coords_f([&E](T val, const auto& d, const auto& c)->T{
//                     return val*E(std::array<T,1>{c[0]});
//                 },"force_i"); // could be made much simpler with abstract iterator functions -> the x axis is the same -> the coordinate is the same;

// upwind
    auto x_adv_1 = stencil<T,I,dim<'x','v'>>(1, {-1.,1},{{-1,0},{0,0}}).get_delta_coords_f([](T val, const auto& d, const auto& c)->T{return c[1]>0?val/d[0]*c[1]:0;},"momentum_1");
    auto x_adv_2 = stencil<T,I,dim<'x','v'>>(1, {-1.,1},{{0,0},{1,0}}).get_delta_coords_f([](T val, const auto& d, const auto& c)->T{return c[1]<0?-val/d[0]*c[1]:0;},"momentum_2");
    auto x_adv = x_adv_1+x_adv_2;
auto v_adv_e_1 = stencil<T,I,dim<'x','v'>>(1,{-1.,1},{{0,-1},{0,0}})
                .get_delta_coords_f([&E](T val, const auto& d, const auto& c)->T{
                    auto tmp = E(std::array<T,1>{c[0]});
                    return E(std::array<T,1>{c[0]})<0?val*E(std::array<T,1>{c[0]}):0;
                },"force_e_p"); // could be made much simpler with abstract iterator functions -> the x axis is the same -> the coordinate is the same;
auto v_adv_e_2 = stencil<T,I,dim<'x','v'>>(1,{-1.,1},{{0,0},{0,1}})
                .get_delta_coords_f([&E](T val, const auto& d, const auto& c)->T{
                    auto tmp = E(std::array<T,1>{c[0]});
                    return E(std::array<T,1>{c[0]})>0?-val*E(std::array<T,1>{c[0]}):0;
                },"force_e_n"); // could be made much simpler with abstract iterator functions -> the x axis is the same -> the coordinate is the same;
auto v_adv_e   = (-.5*q_e/m_e)*(v_adv_e_1+v_adv_e_2);
auto v_adv_i = stencil<T,I,dim<'x','v'>>(.5*q_e/m_i,{-1.,1},{{0,-1},{0,1}})
                .get_delta_coords_f([&E](T val, const auto& d, const auto& c)->T{
                    return val*E(std::array<T,1>{c[0]});
                },"force_i"); // could be made much simpler with abstract iterator functions -> the x axis is the same -> the coordinate is the same;



//collterm
auto C_ei_p1 = stencil<T,I,dim<'x','v'>>(.5, {-1.,1},{{0,-1},{0,1}})
                .get_delta_coords_f([](T val, const auto& d, const auto& c)->T{
                    return val/d[0]*(1/std::abs(c[1])-sgn(c[1])/std::pow(c[1],2));
                },"C_ei_1");

auto C_ei_p2 = stencil<T,I,dim<'x','v'>>(1, {1},{{0},{0}})
                .get_coords_f([](T val, const auto& c)->T{
                    return val*(1/std::pow(std::abs(c[1]),3)-sgn(c[1])/std::pow(c[1],2));
                },"C_ei_2");

auto C_ei = -Gamma_i*(C_ei_p1+C_ei_p2).get_coords_f([&ni](T val, const auto& c)->T{return val*ni(std::array{c[0]});});

auto vlasov_i = (x_adv.unit_stencil() + (x_adv + v_adv_i)).adjust_to_grid(i_space);
auto vlasov_e = (x_adv.unit_stencil() + (x_adv + v_adv_e - C_ei)).adjust_to_grid(e_space);

// e_space.update_boundaries();
// debug::msg2 << e_space << std::endl;
// decltype(e_space) out = vlasov_e*e_space;
// out = vlasov_e * out;
// debug::msg2 << (out) << std::endl;
// return 0;

auto& tmp= E;
auto tmp_i = i_space;
auto tmp_e = e_space;

//     debug::msg2 << Gamma_i << std::endl<< C_ei <<std::endl;
//     debug::msg2 << vlasov_e <<std::endl;
    EXPORT::netcdf(e_space,"e_space.nc").write_grid(e_space);
    auto iterations = gmres(vlasov_e,e_space,1e2)(tmp_e);

    auto diff = tmp_e;
    diff.storage = tmp_e.storage-e_space.storage;

//     debug::msg2 << E << std::endl;
//     debug::msg2 << e_space << std::endl;
//     debug::msg2 << tmp_e << std::endl;
//     debug::msg2 << diff << std::endl;
    debug::msg2 << "iterations: " << iterations << std::endl;
    EXPORT::netcdf(tmp_e,"tmp_e.nc").write_grid(tmp_e);

    return 0;


for (size_t i=0; i<10; ++i){
    debug::msg2 << gmres(vlasov_i,i_space,1e4)(tmp_i) << std::endl;
    debug::msg2 << gmres(vlasov_e,e_space,1e4)(tmp_e) << std::endl;
    ne=tmp_e.template reduction<reduction::SUM,dim<'x'>,basic_iterator>([](const auto& c)->T {return  *c * c.deltas[1];});
    ni=tmp_i.template reduction<reduction::SUM,dim<'x'>,basic_iterator>([](const auto& c)->T {return  *c * c.deltas[1];});
    for(size_t i=0; i<q_space.storage.size(); ++i){
        q_space.storage[i] = -q_e*ne.storage[i]+q_e*ni.storage[i];
    }

    q_space.update_boundaries();
    debug::msg2 << gmres(dx,q_space)(E) << std::endl;
    i_space = tmp_i;
    e_space = tmp_e;
    debug::msg2 << "e_space" << std::endl << e_space.print_all([](T _in)->T {return std::log10(_in/n_e_0);}) << std::endl;
    debug::msg2 << &(tmp.storage[4]) << " " << &(E.storage[4]) << std::endl;
}

debug::msg2 << "e_space" << e_space << std::endl;
debug::msg2 << "i_space" << i_space << std::endl;
debug::msg2 << "electron density" << ne << std::endl;
debug::msg2 << "ion density" << ni << std::endl;
debug::msg2 << "net charge" << q_space << std::endl;
debug::msg2 << "E" << E << std::endl;
// debug::msg2 << "E" << std::endl << E.print_all([](T _in)->T {return std::log(_in+1000);}) << std::endl;
}
