/**
 * @file very_simple_test.cpp
 * @Author Markus Gasteiger
 * @date August, 2018
 * @brief This shall be an example of a very simple test case. Namely, a time independent 1+1 electron only system. There is a given force, supplied by the electric field $E(x)$. Therefor, the resulting differential equation is of the form $$v_x\partial_x f +\frac{q\cdot E(x)}{m_e}\partial_v f = 0 $$.
 *
 */

#include <iostream>
#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"
#include "stencil_func.h"
#include "gmres.h"
#include "constants.h"
#include "export.h"

#include "richardson.h"

using T = double;
using I = long;


T K = 300;
inline T effective_field(T x){
//     return 0;
    T v_thermal = v_thermal_K(K);
    T output = 4e-1*q_e/m_e*std::pow(v_thermal,-2);
    return output;
//     return -output*(std::cos(10*M_PI*x));
    return output * (x<.5?(.5-2*x):(2*x-1.5) +.5);
}


boundary::boundary<T,I,std::array<T,2>> HN(boundary::btype::NEUMANN,1,0.);
boundary::boundary<T,I,std::array<T,2>> HD(boundary::btype::DIRICHLET,1,0);
boundary::boundary<T,I,std::array<T,2>> HG(boundary::btype::DIRICHLET,1,[](const auto& c)->T{return maxwellian(c[1],1.);});
boundary::boundary<T,I,std::array<T,2>> HGbeg(boundary::btype::DIRICHLET,1,maxwellian(0.,1.));
boundary::boundary<T,I,std::array<T,2>> HGend(boundary::btype::DIRICHLET,1,maxwellian(5.,1.));
boundary::boundary<T,I,std::array<T,2>> P(boundary::btype::PERIODIC,1);
// boundary::boundary<T,I,std::array<T,2>> M2(boundary::btype::MIRROR,1);

auto xspace = make_axis<T,I,'x'>(0.,1,400,HG,HG);
// auto vspace = make_axis<T,I,'v'>(0,5.,300,HG,HG);
// auto vspace = make_axis<T,I,'v'>(-5.,5.,600,HG,HG);
auto vspace = make_axis<T,I,'v'>(-5.,5.,600,P,P);
auto space = make_grid<T,I,false>(xspace,vspace);

//centered
int main(){
    T v_thermal = v_thermal_K(K);
T time_step = xspace.delta/std::max(std::abs(vspace.max),std::abs(vspace.min));
debug::output << time_step << std::endl;
for(size_t i=0; i<xspace.points; ++i){
    time_step=std::min(time_step,std::abs(vspace.delta/effective_field(xspace.get_coord(i)))/2);
}
debug::output << time_step << std::endl;

// auto adv_x = stncl<dim<'x','v'>> (-time_step,{-1.,1.},{{-1,0},{0,0}}).get_delta_coords_f([](T val, const auto& deltas, const auto& coords){ return val/deltas[0]*coords[1];});
auto adv_x_a = stncl<dim<'x','v'>> (-time_step,{-1.,1.},{{-1,0},{0,0}}).get_delta_coords_f([](T val, const auto& deltas, const auto& coords){ return coords[1]>0?val/deltas[0]*coords[1]:0;});
auto adv_x_b = stncl<dim<'x','v'>> (-time_step,{-1.,1.},{{0,0},{1,0}}).get_delta_coords_f([](T val, const auto& deltas, const auto& coords){ return coords[1]<0?val/deltas[0]*coords[1]:0;});
// auto adv_x = stncl<dim<'x','v'>> (-time_step/2.,{-1.,1.},{{-1,1},{0,0}}).get_delta_coords_f([](T val, const auto& deltas, const auto& coords){ return val/deltas[0]*coords[1];});
auto adv_x = adv_x_a+adv_x_b;

auto adv_v_o1a = stncl<dim<'x','v'>> (-time_step,{-1.,1.},{{0,0},{0,1}})
        .get_delta_coords_f([](T val, const auto& deltas, const auto& coords){
            T tmp = 1/deltas[1]*effective_field(coords[0]);
            return tmp<0?val*tmp:0;
        });
auto adv_v_o1b = stncl<dim<'x','v'>> (-time_step,{-1.,1.},{{0,-1},{0,0}})
        .get_delta_coords_f([](T val, const auto& deltas, const auto& coords){
            T tmp = 1/deltas[1]*effective_field(coords[0]);
            return tmp>0?val*tmp:0;
        });
auto adv_v_o1 =adv_v_o1a+adv_v_o1b;

// auto adv_v_o1 = stncl<dim<'x','v'>> (-time_step/2,{-1.,1.},{{0,-1},{0,1}})
// .get_delta_coords_f([](T val, const auto& deltas, const auto& coords){
//     T tmp = 1/deltas[1]*effective_field(coords[0]);
//     return val*tmp;
// });



space.update_boundaries();

//     auto FD_o1 = adv_v_o1;
//     auto FD_o1 = adv_x;
    auto FD = adv_x+adv_v_o1;

// return 1;

// auto forward_1 = FD;
// auto backward_o1 = (FD_o1.unit_stencil()-FD_o1).adjust_to_grid(space);
auto forward = (FD.unit_stencil()+FD).adjust_to_grid(space);
auto backward = (FD.unit_stencil()-v_thermal*FD).adjust_to_grid(space);

    space.function<false,false,true>([](const auto& ar)->T {return maxwellian(ar[1],1.);});
//     space.function<true,true,true>([](const T* val, const auto& d, const auto& c)->T { return *val *(1+.5*std::cos(20*M_PI*c[0]));});
//     debug::msg2 << forward_1 << std::endl;
//     debug::msg2 << space.print_all() << std::endl;


//     debug::output << effective_field(.1) << std::endl;

// std::cout << space << std::endl;
    auto tmp = space;
    auto output_gmres = space;
    output_gmres.function<true,false,false>([](T* val)->T{return *val*0.9;});

    debug::msg2 << "v_thermal: " << v_thermal_K(K) << std::endl;

    EXPORT::netcdf(space,"very_simple_space.nc").write_grid(space);
//     return 0;


    size_t iterations1 = 0;
    debug::output << "begin gmres" << std::endl;
    iterations1 = gmres(backward,space,2e4,20,1e-5)(output_gmres);
    EXPORT::netcdf(output_gmres,"very_simple_gmres.nc").write_grid(output_gmres);
    debug::output << "gmres iterations: " << iterations1 << std::endl;

return 0;
    for(size_t i=0; i<1e4;){
        for(size_t j=0; j<100; ++i, ++j){
            auto tmp2 = forward*tmp;
            tmp=tmp2;
        }
        debug::output << "\riteration: " << i;
            EXPORT::netcdf(tmp,"very_simple_forward"+std::to_string(i)+".nc").write_grid(tmp);
    }
        debug::output  << std::endl;
        EXPORT::netcdf(tmp,"very_simple_forward.nc").write_grid(tmp);

//     auto diff = space;
//     diff.storage = output_gmres_o1.storage - output_gmres_o3.storage;
//     debug::msg2 << "gmres "<< output_gmres.print_all() << std::endl,


//     debug::msg2 << forward.to_csr(space) << std::endl;
}
