#ifndef stencil_func_h
#define stencil_func_h

#include <type_traits>
#include <string>
#include <sstream>
#include <array>

#include "stencil_proto.h"
#include "grid.h"
#include "optimize.h"
#include "functional_types.h"
#include "helpers.h"
#include "liste.h"
#include "csr.h"


template<typename T, typename I, class names, bool HD = false, bool is_function = false, bool is_deltas = false, bool is_coords = false, class ... f_args>
struct stencil_func : public stencil_proto<T,I,names> {
    friend grid<T,I,names,HD>;
    using stencil_proto<T,I,names>::dim;
    using this_type = stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>;
    constexpr static bool is_HD = HD;

    // constructors
    stencil_func();

    stencil_func(T coeffi, I offset);
    stencil_func(T coeffi, const std::array<I,names::size>& offset);

    stencil_func(T _factor,
                 const std::vector<T>& _coefficients,
                 const std::vector<std::array<I,dim>>& _positions,
                 const std::array<T,dim>& _deltas = std::array<T,dim> {},
                 const std::vector<T>& _stencil_coefficients = std::vector<T>(0),
                 const typename std::vector<this_type>& _stencils = typename std::vector<this_type>(0),
                 const typename std::function<T(T, const f_args& ...)>& _function = function_deflt<names,T,I,f_args...>,
                 const std::string& f_description = default_function_descriptor);

    stencil_func(const grid<T,I,names,HD>& _grid,
                 T _factor, const std::vector<T>& _coefficients, const std::vector<std::array<I,dim>>& _positions,
                 const std::vector<T>& _stencil_coefficients = std::vector<T>(0),
                 const typename std::vector<this_type>& _stencils = typename std::vector<this_type>(0),
                 const typename std::function<T(T, const f_args& ...)>& _function = function_deflt<names,T,I,f_args...>,
                 const std::string& f_description = default_function_descriptor );

    template<bool is_function_in, bool is_deltas_in, bool is_coords_in, class ... f_args_in>
    stencil_func(const stencil_func<T,I,names,HD,is_function_in,is_deltas_in,is_coords_in, f_args_in...>& _st);

    explicit stencil_func(const stencil_proto<T,I,names>& _st);

// data
    const static bool use_function=is_function || is_deltas || is_coords;
    const static bool use_deltas=is_deltas;
    const static bool use_coords=is_coords;

    const std::array<T,dim> deltas;
    const std::vector<T> stencil_coefficients;
    const std::vector<this_type> stencils;

    const std::function<T(T, const f_args&...)> std_function;
    const std::string function_descriptor;
    template<typename ITER = typename std::vector<T>::const_iterator>
    inline T fun_fun(ITER, const f_args& ...) const;
    template<typename ITER = typename std::vector<T>::const_iterator>
    inline T function(ITER, const f_args& ...) const;


    this_type& operator=(const this_type& _s1);

    stencil_func<T,I,names,HD,true,is_deltas,is_coords,f_args...> get_simple_f( const std::function<T(T)>& _in, const std::string& descriptor = random_string() ){return get_functional<false,false>(_in, descriptor);}

    typename decide_stencil_func_type<names,T,I,true,true,is_coords,HD,stencil_func>::type get_delta_f(const std::function<T(T,const std::array<T,dim>&)>& _in, const std::string& descriptor = random_string() ){return get_functional<true,false,std::array<T,dim>>(_in, descriptor);}

    typename decide_stencil_func_type<names,T,I,true,is_deltas,true,HD,stencil_func>::type get_coords_f(const std::function<T(T,const std::array<T,dim>&)>& _in, const std::string& descriptor = random_string() ){ return get_functional<false,true,std::array<T,dim>>(_in, descriptor);}

    stencil_func<T,I,names,HD,true,true,true,std::array<T,dim>,std::array<T,dim>> get_delta_coords_f(const std::function<T(T, const std::array<T,dim>&, const std::array<T,dim>&)>& _in, const std::string& descriptor = random_string() ){ return get_functional<true,true,std::array<T,dim>, std::array<T,dim>>(_in, descriptor);}

    template<bool HD_new>
    typename decide_stencil_func_type<names,T,I,is_function,is_deltas&&!HD_new,is_coords,HD_new,stencil_func>::type adjust_to_grid(const grid<T,I,names,HD_new>&) const;

    void run_on_grid(grid<T,I,names,HD>& _in, grid<T,I,names,HD>& out) const;
    void run_on_grid(grid<T,I,names,HD>& _grid) const;

    template<bool UPDATE=true>
    grid<T,I,names,HD> operator*(grid<T,I,names,HD>& _in) const;
    template<bool cache=true>
    //NOTE: operator/ overloading causes to ambiguity
    grid<T,I,names,HD> direct_inv(grid<T,I,names,HD> output);

    template<typename Ipos = size_t>
    size_t to_csr_line(std::vector<T>&, std::vector<Ipos>&, Ipos offset, const f_args& ...) const;
    template<typename Ipos = size_t>
    size_t to_csr_line(typename std::vector<T>::iterator, typename std::vector<T>::iterator, typename std::vector<Ipos>::iterator, typename std::vector<Ipos>::iterator, Ipos offset, const f_args& ... args) const;
    csr<T,size_t> to_csr(grid<T,I,names,HD>& _in) const;

    template<bool linear_pos=false>
    std::string print(std::string indent = "") const;

    template<class names_new, bool HD_new, bool is_function_new, bool is_deltas_new, bool is_coords_new, class ... f_args_new>
    auto kron(const stencil_func<T,I,names_new,HD_new,is_function_new,is_deltas_new,is_coords_new,f_args_new...>& _in);

    template<bool adjust_vars=true>
    stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> shift(const std::array<I,dim>& _shift) const;
    template<bool adjust_vars=true>
    stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> shift(const std::vector<T>& coeffi, const std::vector<std::array<I,dim>>& _shift) const;


    auto zero_stencil();
    auto unit_stencil();

    template<bool is_upper, class new_names, bool new_HD>
    typename decide_stencil_func_type<new_names,T,I,is_function,is_deltas,is_coords,HD&&new_HD,stencil_func>::type expand_with(const std::array<I,(new_names::size-names::size)>& array, const std::array<T,(new_names::size-names::size)>& _new_deltas) const;

protected:
    template<bool uses_deltas, bool uses_coords, class ... out_f_args>
    inline typename decide_stencil_func_type<names,T,I,true,uses_deltas||is_deltas,uses_coords || is_coords,HD,stencil_func>::type get_functional(const std::function<T(T,const out_f_args&...)>&, const std::string& descriptor = random_string() );

    std::vector<I> uniq_pos() const;

    csr<T,size_t> inversion_cache_down, inversion_cache_up;
    template<bool BOUNDARIES=false, typename Ipos, typename ITER>
    inline bool get_row(std::vector<Ipos>& _positions, std::vector<T>& _values, const Ipos& counter, const ITER& _gr_iter);
    size_t generate_inversion_cache(grid<T,I,names,HD>& _grid);
    inline void invert_with_cache(grid<T,I,names,HD>& _grid);
    inline void invert_no_cache(grid<T,I,names,HD>& _grid);
};

// Stencil logic
template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
bool operator==(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s1, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s2);

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
void operator*=(grid<T,I,names,HD>& _grid, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s){
//     debug::msg3 << _s.props.strides << _grid.props.strides << std::endl;
    assert(_s.props.strides == _grid.props.strides);
    _s.run_on_grid(_grid);
}

#include "stencil_func.hpp"
#include "stencil_func_multiply.hpp"
#include "stencil_func_helpers.h"
#include "stencil_func_add.hpp"
#include "stencil_func_csr.hpp"
#include "stencil_func_direct_inv.hpp"
#endif
