#ifndef stencil_func_helpers_hpp
#define stencil_func_helpers_hpp

#include "stencil_func.h"

template<int dim, typename T, typename I>
std::array<std::string,dim> mix_stencil_names(const std::array<std::string,dim>& s1, const std::array<std::string,dim>& s2){
    std::array<std::string,dim> target;
    if(s1 == s2){
        target = s1;
    } else { // add logic to accept empty names
        int l1=0, l2=0;
        for(int i=0; i<dim; ++i){
            l1+=s1[i].size();
            l2+=s2[i].size();
        }
        if(l1==0 && l2>0){
            target = s2;
        } else if(l1>0 && l2==0){
            target = s1;
        } else {
            std::invalid_argument("Dimension mismatch.");
        }
    }
    return target;
}

template<typename T, typename I, typename ST>
void add_uniq_stencil_points(std::vector<T>& target_coefficients, std::vector<ST>& target_positions, typename std::vector<T>::const_iterator source_coefficients_begin, typename std::vector<T>::const_iterator source_coefficients_end, typename std::vector<ST>::const_iterator source_positions_begin, typename std::vector<ST>::const_iterator source_positions_end, const T& factor){

    T tmp_value;
    auto iter_coeffi = target_coefficients.begin();
    auto iter_pos = target_positions.begin();
    auto iter_coeffi_s = source_coefficients_begin;
    auto iter_pos_s = source_positions_begin;

    auto inserter_coeffi = target_coefficients.end();
    auto inserter_pos = target_positions.end();

    for(; iter_pos_s!=source_positions_end; ++iter_pos_s, ++iter_coeffi_s) {

        iter_pos = target_positions.begin();
        iter_coeffi = target_coefficients.begin();
        inserter_coeffi = target_coefficients.end();
        inserter_pos = target_positions.end();

        for(; iter_pos!=target_positions.end(); ++iter_pos, ++ iter_coeffi){
            if (*iter_pos==*iter_pos_s) {
                tmp_value=*iter_coeffi+*iter_coeffi_s*factor;

//                 debug::msg2 << tmp_value << std::endl;
                if(is_about_zero(tmp_value)){
                    target_positions.erase(iter_pos);
                    target_coefficients.erase(iter_coeffi);
                } else {
                    *iter_coeffi=tmp_value;
                }
                goto label_end;

            } else {
                // to sort the entries at the same time if an comparison operator exists.
                if constexpr (std::is_arithmetic<ST>::value || std::is_pointer<ST>::value) {
                    if ( *iter_pos > *iter_pos_s ) {
                        inserter_coeffi = iter_coeffi;
                        inserter_pos = iter_pos;
                        break; // some measure of order is assumed
                    } else {
//                         assert(false && "Unordered positions.");
                    }
                }
            }
        }

        target_positions.insert(inserter_pos,*iter_pos_s);
        target_coefficients.insert(inserter_coeffi,*iter_coeffi_s * factor);
        label_end: ;
    }
}

template<typename T, typename I, typename ST>
void add_uniq_stencil_points(std::vector<T>& target_coefficients, std::vector<ST>& target_positions, const std::vector<T>& source_coefficients, const std::vector<ST>& source_positions, const T& factor = 1){
    assert(target_coefficients.size() == target_positions.size());
    assert(source_coefficients.size() == source_positions.size());

    if ( target_coefficients.empty() && source_coefficients.empty() ){
        return;
    }

    add_uniq_stencil_points<T,I,ST>(target_coefficients, target_positions, source_coefficients.cbegin(), source_coefficients.cend(), source_positions.cbegin(), source_positions.cend(), factor);
}

template<typename T, typename I, typename ST>
bool same_stencil_points(const T& _f1, const std::vector<T>& target_coefficients, const std::vector<ST>& target_positions, const T& _f2, const std::vector<T>& source_coefficients, const std::vector<ST>& source_positions){
    //NOTE: this implementation assumes that every position entry is unique in its vector.

    typename std::vector<T>::const_iterator iter_coeffi = target_coefficients.cbegin();
    typename std::vector<ST>::const_iterator iter_pos = target_positions.cbegin();
    typename std::vector<T>::const_iterator iter_coeffi_s = source_coefficients.begin();
    typename std::vector<ST>::const_iterator iter_pos_s = source_positions.begin();

    for(; iter_pos_s!=source_positions.cend(); ++iter_pos_s, ++iter_coeffi_s){
        iter_pos = target_positions.begin();
        iter_coeffi = target_coefficients.begin();
        for(; iter_pos!=target_positions.end(); ++iter_pos, ++ iter_coeffi){
            if(*iter_pos==*iter_pos_s){
                if(is_about_zero(_f1 * *iter_coeffi - _f2 * *iter_coeffi_s)){
                    goto label_end;
                }
            }
        }
        return false;
        label_end: ;
    }
    return true;
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
bool operator==(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s1, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s2){
    //NOTE: Can not compare functions. Hence, this functions is quite impaired.

    if ( !(_s1.props == _s2.props) ) { return false;}

    if(_s1.function_descriptor != _s2.function_descriptor) return false; // But comparing the descriptor works. Then you only need to hope, that they are unique.

    if(! same_stencil_points<T,I>(_s1.factor, _s1.coefficients, _s1.positions, _s2.factor, _s2.coefficients,_s2.positions)) return false;
    if(! same_stencil_points<T,I>(_s1.factor, _s1.stencil_coefficients, _s1.stencils, _s2.factor, _s2.stencil_coefficients,_s2.stencils)) return false;

    return true;
}

template<int dim, typename T, typename I>
std::array<I,dim> compare_strides(const std::array<I,dim>& _s1, const std::array<I,dim>& _s2){
    if ( _s2 == std::array<I,dim> {} ){ return _s1;}
    if ( _s1 == std::array<I,dim> {} ){ return _s2;}
    if ( _s1 != _s2 ) {
        std::invalid_argument("Strides must be identical."); // This is just precautionary.
    }
    return _s1;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
void optimize(stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _in){
    std::vector<T> co = _in.coefficients;
    std::vector<I> lin = _in.linear_positions;
    std::vector<std::array<I,names::size>> pos = _in.positions;

    std::vector<stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>> stenc = _in.stencils;
    for(auto& el: stenc){
        optimize(el);
    }

    sort_v<names::size,T,I>(lin, co, pos);

    _in = stencil_func<T,I,names,HD,is_function,HD,is_coords,f_args...>(_in.strides, _in.factor, co, pos, _in.stencil_coefficients, stenc, _in.std_function);
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
std::ostream& operator<<(std::ostream& os, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _input){
    os << "\n" << std::string(8*(names::size+2),'#') << "\n" << _input.print() << "\n" << std::string(8*(names::size+2),'#');
    return os;
}



#endif
