#define BOOST_TEST_MODULE test_stencil
#include <boost/test/unit_test.hpp>

#include <cmath>
#include <iostream>
#include <string>
#include <limits>

#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"
#include "export.h"

using T = double;
using I = long;


T tolerance = 10*std::numeric_limits<T>::epsilon();

BOOST_AUTO_TEST_SUITE( test_matrix , * boost::unit_test::tolerance(tolerance))

using dm = dim<'x','y','z'>;
using st = stencil<T,I,dm,true>;


T dirichlet1 = 0.33;
T dirichlet2 = 1.57;
T neumann1 = 2.57;

boundary::boundary<T,I,std::array<T,3>> test1(boundary::btype::NEUMANN,2,[](const std::array<T,3>&){return 1.1;});
//\TODO: The boundary functions currently have to be of the same type. This should be more general, i.e. the most general, suitable type should be used.
boundary::boundary<T,I,std::array<T,3>> test2(boundary::btype::OPEN,0);
boundary::boundary<T,I,std::array<T,3>> test3(boundary::btype::DIRICHLET,1,[](const std::array<T,3>&){return 1.1;});

auto axis_x = make_axis<T,I,'x'>(0., 1.,100,test2,test1);
auto axis_y = make_axis<T,I,'y'>(0., 1.,101,test1,test3);
auto axis_z = make_axis<T,I,'z'>(0., 1.,103,test2,test3);

auto grid = make_grid<T,I>(axis_x, axis_y, axis_z);

BOOST_AUTO_TEST_CASE( init ){
    stncl<dim<'x'>> stl(1.,-1);
    stncl<dim<'x'>> stc(-2.,0);
    stncl<dim<'x'>> str(1.,1);
    auto t1 = stl+stc+str;
    t1*= 13.33;
    BOOST_CHECK((t1/2).factor==13.33/2.);
    debug::msg3 << (t1/2).factor << std::endl;

    stncl<dim<'x','y','z'>> st1(1,{{0,0,0}});
    stncl<dim<'x','y','z'>> st2(1.,{{0,-1,0}});
    stncl<dim<'x','y','z'>> st3 = stncl<dim<'x','y','z'>>(0.5,{1,-2,1},{{-1,0,0},{0,0,0},{1,0,0}})/300 - (st1+st2)/10;
    BOOST_CHECK_EQUAL(1/600.,st3.factor);
    BOOST_CHECK(st3.coefficients==std::vector<T>({1.,-62.,1.,-60}));
    debug::msg3 << st3 << std::endl;
}

BOOST_AUTO_TEST_CASE( kronecker ){

    stncl<dim<'x'>> stx(1.,{1,-2,1},{{-1},{0},{1}});
    stncl<dim<'y'>> sty(1.,{1,-2,1},{{-1},{0},{1}});

    auto laplace2d = stx.kron(sty.unit_stencil()) + stx.unit_stencil().kron(sty);
    debug::msg3 << laplace2d << std::endl;
    BOOST_CHECK(laplace2d.coefficients==std::vector<T>({1.,-4.,1.,1.,1.}));
}

    stencil<T,I,dim<'x'>,true> stx(1.,{1,-2,1},{{-1},{0},{1}});
    stencil<T,I,dim<'y'>,true> sty(1.,{1,-2,1},{{-1},{0},{1}});
    auto laplace2d = stx.kron(sty.unit_stencil()) + stx.unit_stencil().kron(sty);
BOOST_AUTO_TEST_CASE( functionalizing ){
    // Only idea to check function: look at printout and some lines down: actually run them

    debug::msg3 << laplace2d << std::endl;

    auto laplace2d_f1 = laplace2d.get_coords_f([](T in, const std::array<T,2>& coords){return in * coords[0];}, "R·x");
    auto laplace2d_f2 = laplace2d.get_delta_f([](T in, const std::array<T,2>& deltas){return in / deltas[1]/ deltas[1];}, "R/dy²");
    auto laplace2d_f = laplace2d_f1+2*laplace2d_f2;
    debug::msg3 << laplace2d_f << std::endl;
}


boundary::boundary<T,I,std::array<T,2>> test_homog_nbc(boundary::btype::NEUMANN,1,[](const std::array<T,2>&){return 0.;});
boundary::boundary<T,I,std::array<T,2>> test_dbc(boundary::btype::DIRICHLET,1,[](const std::array<T,2>&){return 0.;});

auto axis_1 = make_axis<T,I,'x'>(0., 1.,10,test_homog_nbc,test_homog_nbc);
auto axis_2 = make_axis<T,I,'y'>(0., 1.,11,test_dbc,test_dbc);
auto grid2 = make_grid<T,I,false>(axis_1, axis_2);
// ::grid<T, I, dim<'x','y'>, true> grid3();
stencil_func<T,I,dim<'x','y'>,true,false,false,false> ttest(grid2, 1., std::vector<T>({}), std::vector<std::array<I,2>>({}));


BOOST_AUTO_TEST_CASE( adjust_stencil_to_grid ){
debug::msg3 << laplace2d << std::endl;
debug::msg3 << grid2.axes[0]->template get_coord(2) << std::endl;
debug::msg3 << grid2 << std::endl;
    auto laplace2d_a = laplace2d.adjust_to_grid(grid2);
debug::msg3 << grid2.print() << std::endl;
}


auto laplace2d_a = laplace2d.adjust_to_grid(grid2);
auto laplace2d_a_f = laplace2d_a.get_coords_f([](T in, const std::array<T,2>& coords){return in * coords[0];}, "R·x");
auto laplace2d_a_f2 = laplace2d_a.get_delta_f([](T in, const std::array<T,2>& coords){return in / coords[0];}, "R/dx");
BOOST_AUTO_TEST_CASE( run_stencil_on_grid ){
    grid2.function<false,false,true>([](const std::array<T,2>& ar){ return ar[1]+ar[0]/10.;});
    debug::msg3 << grid2 << std::endl;
//
    auto output = (laplace2d_a.adjust_to_grid(grid2))*grid2;
    auto output2 = (laplace2d_a_f.adjust_to_grid(grid2))*output;
    auto output3 = (laplace2d_a_f2.adjust_to_grid(grid2))*output2;
//
    debug::msg3 << output << std::endl;
}

stencil<T,I,dim<'x'>,true> adv_x(1.,{1.},{{-1}});
stencil<T,I,dim<'y'>,true> adv_y(1.,{1.},{{-1}});
auto adv = (adv_x.kron(adv_y)).adjust_to_grid(grid2);
BOOST_AUTO_TEST_CASE( stencil_correct ){
    debug::msg3 << adv << std::endl;
    debug::msg3 << grid2.print_all() << std::endl;
    debug::msg3 << "here" << std::endl;
    grid2.function<false,false,true>([](const std::array<T,2>& ar){ return ar[1]+ar[0]/10.;});
//     grid2.update_boundaries();
    auto res = adv*grid2;
//     res.update_boundaries();
    debug::msg3 << grid2.print_all() << std::endl;
    debug::msg3 << res.print_all() << std::endl;

    BOOST_CHECK_EQUAL((res[ std::array<size_t,2> {4,1} ]),0.033333333333333333);
}

BOOST_AUTO_TEST_CASE( laplacian_2d ){
//     debug_level tmp_l = debug::level;
//     debug::level = debug_level::FULL;

    debug::msg3 << "Test a 2d laplacian with delta peak initial data." << std::endl;
//     debug::msg3 << laplace2d << std::endl;
    boundary::boundary<T,I,std::array<T,2>> mirror_bc(boundary::btype::PERIODIC,1);
    auto x_axis = make_axis<T,I,'x'>(-.5,.5,15,mirror_bc,mirror_bc);
    auto y_axis = make_axis<T,I,'y'>(-.5,.5,15,mirror_bc,mirror_bc);
    auto square_grid = make_grid<T,I,false>(x_axis,y_axis);
    square_grid.function<false,false,true>([](const std::array<T,2>& _ar){
        if(std::abs(_ar[0]) < 0.2 &&std::abs(_ar[1])<0.2){
            return 1.;
        } else {
            return 0.;
        }
    });
    debug::msg3 << square_grid.print_all() << std::endl;
    auto lplc2d = (stencil<T,I,dim<'x','y'>,true>(1.,{1.},{{0}}) + laplace2d/10.).adjust_to_grid(square_grid);
    auto adv2 = adv.adjust_to_grid(square_grid);
    BOOST_CHECK(lplc2d.props.strides == square_grid.props.strides);
//     debug::msg3 << std::endl << lplc2d.template print<true>() <<std::endl;

    decltype(square_grid) res = square_grid;
    decltype(square_grid) res_2 = square_grid;
//     debug::msg3 << res.props.strides << lplc2d.props.strides << std::endl;
//     res = lplc2d*square_grid;
    auto output_obj = EXPORT::netcdf<T,I,dim<'x','y'>,true,true>(res,"laplace2d.nc");
    output_obj.write_grid(res);
    for(size_t i=0; i<10; ++i){
        res_2 = lplc2d*res;
        res = res_2;
//         res.update_boundaries();
        output_obj.write_grid(res);
//         debug::msg3 << res.print_all() << std::endl;
//         res= res_tmp;
    }
//     debug::msg3 << res.print_all() << std::endl;
//     debug::msg3 << lplc2d << std::endl;

//     debug::level = tmp_l;
}

BOOST_AUTO_TEST_CASE( csr_generation ){
    boundary::boundary<T,I,std::array<T,2>> mirror_bc(boundary::btype::PERIODIC,1);
    auto x_axis = make_axis<T,I,'x'>(-.5,.5,15,mirror_bc,mirror_bc);
    auto y_axis = make_axis<T,I,'y'>(-.5,.5,15,mirror_bc,mirror_bc);
    auto square_grid = make_grid<T,I,false>(x_axis,y_axis);
    auto lplc2d = laplace2d.adjust_to_grid(square_grid);
//     auto lplc2d = (stencil<T,I,dim<'x','y'>,true>(1.,{1.},{{0}}) + laplace2d/10.).adjust_to_grid(square_grid);

    auto csr_matrix = lplc2d.to_csr(square_grid);

    // the following comparison matrices were only inspected visually for compliance. Hence, not all values have been checked to adhere to the csr form of the 2d Laplacian. But it is assumed to be consistent and sufficient to detect future regressions.
    std::vector<T> compare_A = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,-4,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
    std::vector<size_t> compare_JA = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,1,17,18,19,35,2,18,19,20,36,3,19,20,21,37,4,20,21,22,38,5,21,22,23,39,6,22,23,24,40,7,23,24,25,41,8,24,25,26,42,9,25,26,27,43,10,26,27,28,44,11,27,28,29,45,12,28,29,30,46,13,29,30,31,47,14,30,31,32,48,15,31,32,33,49,33,34,18,34,35,36,52,19,35,36,37,53,20,36,37,38,54,21,37,38,39,55,22,38,39,40,56,23,39,40,41,57,24,40,41,42,58,25,41,42,43,59,26,42,43,44,60,27,43,44,45,61,28,44,45,46,62,29,45,46,47,63,30,46,47,48,64,31,47,48,49,65,32,48,49,50,66,50,51,35,51,52,53,69,36,52,53,54,70,37,53,54,55,71,38,54,55,56,72,39,55,56,57,73,40,56,57,58,74,41,57,58,59,75,42,58,59,60,76,43,59,60,61,77,44,60,61,62,78,45,61,62,63,79,46,62,63,64,80,47,63,64,65,81,48,64,65,66,82,49,65,66,67,83,67,68,52,68,69,70,86,53,69,70,71,87,54,70,71,72,88,55,71,72,73,89,56,72,73,74,90,57,73,74,75,91,58,74,75,76,92,59,75,76,77,93,60,76,77,78,94,61,77,78,79,95,62,78,79,80,96,63,79,80,81,97,64,80,81,82,98,65,81,82,83,99,66,82,83,84,100,84,85,69,85,86,87,103,70,86,87,88,104,71,87,88,89,105,72,88,89,90,106,73,89,90,91,107,74,90,91,92,108,75,91,92,93,109,76,92,93,94,110,77,93,94,95,111,78,94,95,96,112,79,95,96,97,113,80,96,97,98,114,81,97,98,99,115,82,98,99,100,116,83,99,100,101,117,101,102,86,102,103,104,120,87,103,104,105,121,88,104,105,106,122,89,105,106,107,123,90,106,107,108,124,91,107,108,109,125,92,108,109,110,126,93,109,110,111,127,94,110,111,112,128,95,111,112,113,129,96,112,113,114,130,97,113,114,115,131,98,114,115,116,132,99,115,116,117,133,100,116,117,118,134,118,119,103,119,120,121,137,104,120,121,122,138,105,121,122,123,139,106,122,123,124,140,107,123,124,125,141,108,124,125,126,142,109,125,126,127,143,110,126,127,128,144,111,127,128,129,145,112,128,129,130,146,113,129,130,131,147,114,130,131,132,148,115,131,132,133,149,116,132,133,134,150,117,133,134,135,151,135,136,120,136,137,138,154,121,137,138,139,155,122,138,139,140,156,123,139,140,141,157,124,140,141,142,158,125,141,142,143,159,126,142,143,144,160,127,143,144,145,161,128,144,145,146,162,129,145,146,147,163,130,146,147,148,164,131,147,148,149,165,132,148,149,150,166,133,149,150,151,167,134,150,151,152,168,152,153,137,153,154,155,171,138,154,155,156,172,139,155,156,157,173,140,156,157,158,174,141,157,158,159,175,142,158,159,160,176,143,159,160,161,177,144,160,161,162,178,145,161,162,163,179,146,162,163,164,180,147,163,164,165,181,148,164,165,166,182,149,165,166,167,183,150,166,167,168,184,151,167,168,169,185,169,170,154,170,171,172,188,155,171,172,173,189,156,172,173,174,190,157,173,174,175,191,158,174,175,176,192,159,175,176,177,193,160,176,177,178,194,161,177,178,179,195,162,178,179,180,196,163,179,180,181,197,164,180,181,182,198,165,181,182,183,199,166,182,183,184,200,167,183,184,185,201,168,184,185,186,202,186,187,171,187,188,189,205,172,188,189,190,206,173,189,190,191,207,174,190,191,192,208,175,191,192,193,209,176,192,193,194,210,177,193,194,195,211,178,194,195,196,212,179,195,196,197,213,180,196,197,198,214,181,197,198,199,215,182,198,199,200,216,183,199,200,201,217,184,200,201,202,218,185,201,202,203,219,203,204,188,204,205,206,222,189,205,206,207,223,190,206,207,208,224,191,207,208,209,225,192,208,209,210,226,193,209,210,211,227,194,210,211,212,228,195,211,212,213,229,196,212,213,214,230,197,213,214,215,231,198,214,215,216,232,199,215,216,217,233,200,216,217,218,234,201,217,218,219,235,202,218,219,220,236,220,221,205,221,222,223,239,206,222,223,224,240,207,223,224,225,241,208,224,225,226,242,209,225,226,227,243,210,226,227,228,244,211,227,228,229,245,212,228,229,230,246,213,229,230,231,247,214,230,231,232,248,215,231,232,233,249,216,232,233,234,250,217,233,234,235,251,218,234,235,236,252,219,235,236,237,253,237,238,222,238,239,240,256,223,239,240,241,257,224,240,241,242,258,225,241,242,243,259,226,242,243,244,260,227,243,244,245,261,228,244,245,246,262,229,245,246,247,263,230,246,247,248,264,231,247,248,249,265,232,248,249,250,266,233,249,250,251,267,234,250,251,252,268,235,251,252,253,269,236,252,253,254,270,254,255,239,255,256,257,273,240,256,257,258,274,241,257,258,259,275,242,258,259,260,276,243,259,260,261,277,244,260,261,262,278,245,261,262,263,279,246,262,263,264,280,247,263,264,265,281,248,264,265,266,282,249,265,266,267,283,250,266,267,268,284,251,267,268,269,285,252,268,269,270,286,253,269,270,271,287,271,272,273,274,275,276,277,278,279,280,281,282,283,284,285,286,287,288};
    std::vector<size_t> compare_IA = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,23,28,33,38,43,48,53,58,63,68,73,78,83,88,93,94,95,100,105,110,115,120,125,130,135,140,145,150,155,160,165,170,171,172,177,182,187,192,197,202,207,212,217,222,227,232,237,242,247,248,249,254,259,264,269,274,279,284,289,294,299,304,309,314,319,324,325,326,331,336,341,346,351,356,361,366,371,376,381,386,391,396,401,402,403,408,413,418,423,428,433,438,443,448,453,458,463,468,473,478,479,480,485,490,495,500,505,510,515,520,525,530,535,540,545,550,555,556,557,562,567,572,577,582,587,592,597,602,607,612,617,622,627,632,633,634,639,644,649,654,659,664,669,674,679,684,689,694,699,704,709,710,711,716,721,726,731,736,741,746,751,756,761,766,771,776,781,786,787,788,793,798,803,808,813,818,823,828,833,838,843,848,853,858,863,864,865,870,875,880,885,890,895,900,905,910,915,920,925,930,935,940,941,942,947,952,957,962,967,972,977,982,987,992,997,1002,1007,1012,1017,1018,1019,1024,1029,1034,1039,1044,1049,1054,1059,1064,1069,1074,1079,1084,1089,1094,1095,1096,1101,1106,1111,1116,1121,1126,1131,1136,1141,1146,1151,1156,1161,1166,1171,1172,1173,1174,1175,1176,1177,1178,1179,1180,1181,1182,1183,1184,1185,1186,1187,1188,1189};

    BOOST_CHECK(csr_matrix.A.size() == compare_A.size());
    BOOST_CHECK(csr_matrix.JA.size() == compare_JA.size());
    BOOST_CHECK(csr_matrix.IA.size() == compare_IA.size());
    BOOST_CHECK(csr_matrix.A == compare_A);
    BOOST_CHECK(csr_matrix.JA == compare_JA);
    BOOST_CHECK(csr_matrix.IA == compare_IA);

//     debug::msg3 << csr_matrix.A << std::endl;
//     debug::msg3 << csr_matrix.IA << std::endl;
//     debug::msg3 << csr_matrix.JA << std::endl;
}


BOOST_AUTO_TEST_CASE( matrix_inversion_1d ){
    boundary::boundary<T,I,std::array<T,1>> inversion_bc(boundary::btype::DIRICHLET,1,0);
    auto x_axis = make_axis<T,I,'x',discretization::HOMOGENEOUS>(-.5,.5,3,inversion_bc,inversion_bc);
    auto square_grid = make_grid<T,I,false>(x_axis);
    auto lplc1d = stx.adjust_to_grid(square_grid);

    square_grid.function<false,false,true>([](const std::array<T,1>& ar){ if(is_about_zero(std::abs(ar[0]))){ return 1.; } else { return 0.;} });
    debug::msg3 << square_grid.print_all() << std::endl;
//
    auto intermediate = lplc1d*square_grid;
    debug::msg3 << intermediate.print_all() << std::endl;
//
    auto output = lplc1d.template direct_inv<false>(intermediate);
    debug::msg3 << output.print_all() << std::endl;

    BOOST_CHECK(intermediate.storage != square_grid.storage);
    BOOST_CHECK(output.storage == square_grid.storage);

    lplc1d.direct_inv(intermediate);
    auto output2 = lplc1d.direct_inv(intermediate);
    BOOST_CHECK(output2.storage == square_grid.storage);
    debug::msg2 << lplc1d << square_grid.print_all() << intermediate.print_all() << output2.print_all() << std::endl;
}

BOOST_AUTO_TEST_CASE( matrix_inversion_2d ){
    boundary::boundary<T,I,std::array<T,2>> inversion_bc(boundary::btype::DIRICHLET,1,0);
    auto x_axis = make_axis<T,I,'x'>(-.5,.5,3,inversion_bc,inversion_bc);
    auto y_axis = make_axis<T,I,'y'>(-.5,.5,3,inversion_bc,inversion_bc);
    auto square_grid = make_grid<T,I,false>(x_axis,y_axis);
    auto lplc2d = laplace2d.adjust_to_grid(square_grid);

    square_grid.function<false,false,true>([](const std::array<T,2>& ar){ if(is_about_zero(std::abs(ar[0]-ar[1]))){ return 1.; } else { return 0.;} });
    square_grid.function<false,false,true>([](const std::array<T,2>& ar){ return ar[0]+ar[1]/10.; });
    debug::msg3 << square_grid.print_all() << std::endl;

    auto intermediate = lplc2d*square_grid;
    debug::msg3 << intermediate.print_all() << std::endl;

    auto output = lplc2d.template direct_inv<false>(intermediate);
    debug::msg3 << output.print_all() << std::endl;

    BOOST_CHECK(intermediate.storage != square_grid.storage);
    BOOST_TEST(output.storage == square_grid.storage, boost::test_tools::per_element() );

    lplc2d.direct_inv(intermediate);
    auto output2 = lplc2d.direct_inv(intermediate);
    BOOST_TEST(output2.storage == square_grid.storage, boost::test_tools::per_element() );

    debug::msg3 << square_grid.storage << std::endl;
    debug::msg3 << output.storage << std::endl;

}

BOOST_AUTO_TEST_CASE( matrix_inversion_diagonal_zeroes ){
    //NOTE: In this case we want to test the algorithm with the representation of the matrix [[1,0,0,0,0,0],[1,0,1,0,0,0],[0,1,0,1,0,0],[0,0,1,0,1,0],[0,0,0,1,0,1],[0,0,0,0,0,1]]. That matrix has zeroes for diagonal elements and should therefore make problems for the algorithm.

    boundary::boundary<T,I,std::array<T,1>> inversion_bc(boundary::btype::DIRICHLET,1,0);
    auto x_axis = make_axis<T,I,'x',discretization::HOMOGENEOUS>(-.5,.5,4,inversion_bc,inversion_bc);
    auto square_grid = make_grid<T,I,false>(x_axis);
    stencil<T,I,dim<'x'>,true> mean_x(1.,{1.,1.},{{-1},{1}});
    auto mean = mean_x.adjust_to_grid(square_grid);

    square_grid.function<false,false,true>([](const std::array<T,1>& ar){ if(std::abs(ar[0])<.2){ return 1.; } else { return 0.;} });
    //
    auto tcsr = mean.to_csr(square_grid);
    debug::msg3 << tcsr << std::endl;

    auto intermediate = mean*square_grid;
    debug::msg3 << intermediate.print_all() << std::endl;
    //
    auto output = mean.direct_inv(intermediate);
    debug::msg3 << square_grid.print_all()<< std::endl << intermediate.print_all() << std::endl << output.print_all() << std::endl;

    BOOST_CHECK(intermediate.storage != square_grid.storage);
    BOOST_CHECK(output.storage == square_grid.storage);
}
    //NOTE: what would be relevant additional test for the inversion operation?

BOOST_AUTO_TEST_CASE( stencil_shifting ){
    boundary::boundary<T,I,std::array<T,1>> inversion_bc(boundary::btype::DIRICHLET,1,0);
    auto x_axis = make_axis<T,I,'x',discretization::HOMOGENEOUS>(-.5,.5,4,inversion_bc,inversion_bc);
    auto square_grid = make_grid<T,I,false>(x_axis);

    auto tester = laplace2d.shift(std::array<I,2> {1,-1});

    debug::msg2 << laplace2d << std::endl;
    debug::msg2 << tester << std::endl;

    auto basis1 = stncl<dim<'x'>>(1.,{-1,1},{{-1},{0}}).adjust_to_grid(square_grid);
    auto basis2 = stncl<dim<'x'>>(1.,{-1,1},{{0},{1}}).adjust_to_grid(square_grid);
    auto m1 = basis2 - basis1.get_delta_f([](T v, const std::array<T,1>& c){return v/c[0];});
    auto m2 = basis1 - basis2.get_coords_f([](T v, const std::array<T,1>& c){ return v*c[0];});
    debug::msg2 << m1 << std::endl << "*" << m2 << std::endl << "=" << std::endl;
    debug::msg2 << m1*m2 <<std::endl;
}

BOOST_AUTO_TEST_SUITE_END()
